# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast

from contrast.agent.policy.constants import CURRENT_FINDING_VERSION, TRIGGER_TYPE
from contrast.api.dtm_pb2 import TraceEvent
from contrast.extern import six
from contrast.utils.base64_utils import base64_decode

# These request part values must match the TYPE parameter for source nodes in policy.json, which are
# a best effort at translating the types in this document:
# https://bitbucket.org/contrastsecurity/teamserver/src/862d499227e8eac42b6eb7c6b03f10b7f1556218/teamserver-agent-messages/src/main/java/contrast/agent/messages/finding/trace/EventSourceTypeDTM.java#lines-8
SOURCE_OPTIONS = {
    "parameter": "PARAMETER",
    "cookies": "COOKIE",
    "body": "BODY",
    "full_path": "URI",
    "full_path_info": "URI",
    "raw_uri": "URI",
    "host": "URI",
    "port": "URI",
    "scheme": "OTHER",
    "encoding": "OTHER",
    "files": "MULTIPART_CONTENT_DATA",
    "form": "MULTIPART_FORM_DATA",
    "header": "HEADER",
    "referer_header": "HEADER",
    "header_key": "HEADER_KEY",
    "wsgi.input": "MULTIPART_CONTENT_DATA",
}


def validate_finding(
    finding, rule_id, user_input=None, nondataflow=False, config=False, trigger_arg=0,
):
    assert finding.rule_id == rule_id

    assert finding.version == CURRENT_FINDING_VERSION
    assert len(finding.routes) == 1

    if config:
        return

    if nondataflow:
        assert len(finding.events) == 1
        return

    trigger_event = finding.events[-1]
    assert trigger_event.action == TraceEvent.Action.Value(TRIGGER_TYPE)

    if trigger_event.source == "RETURN":
        trigger_input = base64_decode(trigger_event.ret.value)
    elif trigger_event.source == "OBJ":
        trigger_input = base64_decode(trigger_event.object.value)
    else:
        # assume it's args[0] unless we specify otherwise
        trigger_input = base64_decode(trigger_event.args[trigger_arg].value)

    # original user input should be somehow in the trigger event or we've
    # string-ified a bytes-like obj or a regex pattern object
    assert (
        user_input in trigger_input
        or "io.BytesIO" in trigger_input
        or "_sre.SRE_Pattern" in trigger_input
    )


def validate_finding_in_context(
    request_context,
    rule_id,
    user_input=None,
    nondataflow=False,
    config=False,
    trigger_arg=0,
):
    finding = _get_single_finding_or_fail(
        request_context, rule_id, nondataflow or config
    )

    validate_finding(finding, rule_id, user_input, nondataflow, config, trigger_arg)


def validate_source_finding_in_context(
    request_context, rule_id, source_name, source_map=None
):
    finding = _get_single_finding_or_fail(request_context, rule_id, nondataflow=False)
    assert finding.rule_id == rule_id
    assert_event_sources([finding], source_name, source_map or SOURCE_OPTIONS)


def remove_redos_findings(findings, rule_id):
    """
    Redos rule creates vulnerabilities in lots of places including framework
    code. For now we ignore a redos vuln if we're are testing for redos rule

    NOTE: These redos findings are legitimate and at this time we are reporting them
    to TS. Work will labs will decide if that should continue. We only filter them
    out in these tests to make assertions easy.
    """
    if rule_id == "redos":
        return findings

    return [x for x in findings if x.rule_id != "redos"]


def validate_nondataflow_finding(
    findings, response, mocked_build_finding, rule_id, call_count, num_findings=1
):
    assert response is not None

    filtered_findings = remove_redos_findings(findings, rule_id)
    if len(filtered_findings) < len(findings):
        # there was likely a redos finding in the framework so increment call count
        call_count += 1

    assert mocked_build_finding.called
    assert mocked_build_finding.call_count == call_count

    args, _ = mocked_build_finding.call_args

    assert args[1].name == rule_id

    assert len(filtered_findings) == num_findings

    finding = filtered_findings[0]
    assert len(finding.routes) == 1
    assert len(finding.events) == 1
    assert finding.version == CURRENT_FINDING_VERSION


def validate_provider_finding(mock_send_finding):
    hardcoded_key = mock_send_finding.call_args_list[0][0][0]
    hardcoded_key.rule_id == "hardcoded-key"

    properties = hardcoded_key.properties
    properties["codeSource"] == "SECRET_KEY = [**REDACTED**]"
    properties["name"] == "SECRET_KEY"
    "app/settings.py" in properties["source"]

    hardcoded_pwd = mock_send_finding.call_args_list[1][0][0]
    hardcoded_pwd.rule_id == "hardcoded-password"

    properties = hardcoded_pwd.properties
    properties["codeSource"] == "PASSWORD = [**REDACTED**]"
    properties["name"] == "PASSWO RD"
    "app/settings.py" in properties["source"]


def validate_source_finding(
    findings,
    response,
    mocked_build_finding,
    source_name,
    rule_id,
    num_findings,
    source_map=None,
):
    assert response is not None
    assert mocked_build_finding.called

    args, _ = mocked_build_finding.call_args

    assert args[1].name == rule_id

    findings = remove_redos_findings(findings, rule_id)

    assert len(findings) == num_findings

    assert_event_sources(findings, source_name, source_map or SOURCE_OPTIONS)


def validate_header_sources(findings, source_name):
    rule_id = "cmd-injection"
    findings = remove_redos_findings(findings, rule_id)

    assert len(findings) == 1
    assert findings[0].rule_id == "cmd-injection"
    assert_event_sources(
        findings, source_name, {"header": "HEADER", "header_key": "HEADER_KEY"}
    )


def validate_cookies_sources(findings):
    rule_id = "cmd-injection"
    findings = remove_redos_findings(findings, rule_id)

    assert len(findings) == 1
    assert findings[0].rule_id == rule_id
    assert_event_sources(findings, "cookies", {"cookies": "COOKIE"})


def assert_event_sources(findings, source_name, source_options):
    all_event_sources = _get_event_sources(findings)
    # for now all we care is that the source name appears in TYPE at least once
    assert source_options[source_name] in [source.type for source in all_event_sources]


def _get_single_finding_or_fail(request_context, rule_id, nondataflow):
    """
    Get the single finding associated with the given request context.
    If there are multiple findings, fail the unit test.

    Nondataflow findings, especially crypto-bad-mac, are fairly common in
    various frameworks. For tests where we only care about the dataflow
    findings, this method removes any additional nondataflow findings.
    """
    if nondataflow:
        findings = request_context.activity.findings
    else:
        findings = [
            finding
            for finding in request_context.activity.findings
            if finding.rule_id
            # we may need to expand this list in the future
            not in ("crypto-bad-mac", "crypto-weak-randomness", "session-timeout")
        ]

    findings = remove_redos_findings(findings, rule_id)
    assert len(findings) == 1
    return findings[0]


def _get_event_sources(findings):
    """
    Get all trace event sources from all finding.events

    :param findings: a list of Finding objects
    :return: a list of TraceEventSource objects
    """
    event_sources = []

    for finding in findings:
        for event in finding.events:
            if event.event_sources:
                for trace_event in event.event_sources:
                    event_sources.append(trace_event)

    return event_sources


def _validate(response, tracked_param, mocked_build_finding):
    assert response is not None
    assert param_is_tracked(tracked_param)
    assert mocked_build_finding.called


def param_is_tracked(param_val):
    for props in contrast.STRING_TRACKER.values():
        if props.origin == param_val:
            return True
    return False


def validate_ssrf_finding(trigger, request_context, mocked):
    assert mocked.called
    assert mocked.call_args[0][1].name == "ssrf"


def validate_ssrf_finding_no_mock(trigger, request_context):
    # This case results in multiple findings for reasons that are not entirely understood
    num_findings = (
        2
        if trigger == "urlopen-obj"
        or six.PY2
        and trigger in ["legacy-urlopen", "urlopen-str"]
        else 1
    )

    findings = remove_redos_findings(request_context.activity.findings, "ssrf")
    assert len(findings) == num_findings
    assert findings[0].rule_id == six.u("ssrf")



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast


class InitialMsgsTestMixin(object):
    def test_initial_msgs(self, mocker):
        self.client.get("/", status="*")

        # send_dtm is the last mock mocker created in disable_service_client
        send_dtm_mock = mocker._mocks[-1]

        initial_messages = send_dtm_mock.call_args_list[:2]

        first_msg = initial_messages[0][0][0]
        assert first_msg.app_name == "root"
        assert first_msg.app_language == "Python"
        assert first_msg.app_path == "/app-path"

        agent_startup = first_msg.agent_startup
        app_create = initial_messages[1][0][0].application_create

        assert agent_startup.version == contrast.__version__
        assert agent_startup.server_version == contrast.__version__

        assert agent_startup.server_name == "test_server_name"
        assert agent_startup.server_path == "/test/server/path"
        assert agent_startup.server_tags == "some_test_server_tag"
        assert agent_startup.application_tags == "some_test_application_tag"
        assert agent_startup.library_tags == "some_test_library_tag"
        assert agent_startup.finding_tags == "some_test_assess_tag"

        assert app_create.group == "some_test_group"
        assert app_create.app_version == "some_test_version"
        assert app_create.code == "some_test_application_code"
        assert not app_create.mode.protect
        assert app_create.mode.assess



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.api.settings_pb2 import ProtectionRule


class MalformedHeader(BaseRule):
    """
    Malformed Header Protection rule
    """

    NAME = "malformed-header"

    def __init__(self, settings):
        BaseRule.__init__(self, settings)

    @property
    def mode(self):
        """
        Always block at perimeter
        """
        return ProtectionRule.BLOCK_AT_PERIMETER



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
from contrast.extern.six import PY3


class HardcodedPasswordHolder(object):
    # Should be found
    PASSWORD = "foo"
    PASSKEY = "foo"
    PASSPHRASE = "foo"
    SECRET = "foo"
    PASSWORD_OK = b"foo" if PY3 else bytearray("foo")
    PASSWORD_UNDER = "__foo"

    # Should not be found - names
    PASSWORD_DATE = "foo"
    PASSWORD_FORGOT = "foo"
    PASSWORD_FORM = "foo"
    PASSWORD_ENCODE = "foo"
    PASSWORD_PATTERN = "foo"
    PASSWORD_PREFIX = "foo"
    PASSWORD_PROP = "foo"
    PASSWORD_SUFFIX = "foo"
    PASSWORD_URL = "foo"
    PASSWORD_BASE = "foo"
    PaSsWoRd_BaSe = "foo"
    PASSWORD_FILE = "foo"
    __PASSWORD = "foo"
    KEY_REASSIGN = PASSWORD

    # Should not be found - values
    PASSWORD_DOTS = "foo.bar"
    PASSWORD_UNDERSCORE = "foo_bar"
    PASSWORD_MIXED = "foo.bar_baz"
    PASSWORD_NUMBERS = 12345
    os.environ.setdefault("PASSWORD", "foo")
    PASSWORD = os.environ.get("PASSWORD")



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re
import ast
import textwrap
from inspect import getsource, isclass
from types import FunctionType, MethodType

from contrast.api.dtm_pb2 import RouteCoverage

DEFAULT_ROUTE_METHODS = ("GET", "POST")


class CoverageUtils(object):
    """
    Rules for coverage:

    - new routes on init get count 0
    - new routes after init get count 1 per context
    - routes without a method type specified on init are ['GET', 'POST']
    - a route needs verb, route, url, and count
    - url should be normalized uri
    - route should be the path to the view function (aka controller for Java people)
    - one method type per route

    Example:
          GET /blog/foo/bar - app.blogs.find(request, ) 0
    """

    # 1 or more digits, then a group: either '/' or end of string
    DIGITS_ONLY_PATH_COMPONENT = r"/\d+(/|$)"

    @staticmethod
    def get_normalized_uri(path):
        """
        A best-effort to remove client-specific information from the path.

        Example:
        /user/123456/page/12 -> /user/{n}/page/{n}
        """
        return re.sub(CoverageUtils.DIGITS_ONLY_PATH_COMPONENT, r"/{n}\1", path)

    @staticmethod
    def build_route_coverage(**kwargs):
        route_coverage = RouteCoverage()
        route_coverage.verb = kwargs.get("verb", "GET")
        route_coverage.url = kwargs.get("url", "")
        route_coverage.route = kwargs.get("route", "")

        return route_coverage

    @staticmethod
    def check_for_http_decorator(func):
        """
        Grabs the require_http_methods decorator from a view function call via inspect

        NOTE: this only works for require_http_methods; it does not currently work for require_GET,
        require_POST, require_safe
        """
        method_types = {}

        if func is None:
            return method_types

        def visit_function_def(node):
            for n in node.decorator_list:
                if isinstance(n, ast.Call) and hasattr(n, "func"):
                    node_func = n.func

                    # decorator name which should be require_http_methods
                    if (
                        hasattr(node_func, "id")
                        and node_func.id == "require_http_methods"
                    ):
                        name = (
                            node_func.attr
                            if isinstance(node_func, ast.Attribute)
                            else node_func.id
                        )
                        method_types[name] = [s.s for s in n.args[0].elts]
                        return

        node_iter = ast.NodeVisitor()
        node_iter.visit_FunctionDef = visit_function_def
        node_source = textwrap.dedent(getsource(func))
        node_iter.visit(ast.parse(node_source))

        return method_types

    @staticmethod
    def get_lowest_function_call(func):
        if isclass(func) or func.__closure__ is None:
            return func
        closure = (c.cell_contents for c in func.__closure__)
        return next(
            (c for c in closure if isinstance(c, (FunctionType, MethodType))), None
        )

    @staticmethod
    def build_args_from_function(func):
        """
        Attempts to grab argument names from the function definition.

        Defaults to () if none exist
        If there is no view function, like in the case of a pure WSGI app, then the func will
        be a string like '/sqli' and we just return that.

        func_code is for PY2 and __code__ is for PY3
        """
        method_arg_names = "()"
        if func is not None and hasattr(func, "func_code"):
            method_arg_names = str(
                func.func_code.co_varnames[0 : func.func_code.co_argcount]
            ).replace("'", "")
        elif func is not None and hasattr(func, "__code__"):
            method_arg_names = str(
                func.__code__.co_varnames[0 : func.__code__.co_argcount]
            ).replace("'", "")
        elif isinstance(func, str):
            method_arg_names = func

        return method_arg_names

    @staticmethod
    def build_key(route_id, method):
        return "_".join([route_id, method])



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent.protect.rule.base_rule import BaseRule, UserInput
from contrast.utils.decorators import set_context

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class MethodTampering(BaseRule):
    NAME = "method-tampering"
    USER_INPUT_KEY = UserInput.InputType.Name(UserInput.METHOD)

    def __init__(self, settings):
        BaseRule.__init__(self, settings)

    @set_context("in_postfilter")
    def postfilter(self):
        """
        At postfilter we generate activity if input analysis was found and depending on application response code.

        if response code is either 4xx or 5xx, application was not exploited (only probed) by an unexpected HTTP method.
        If response code is anything else, then an unexpected HTTP method successfully exploited the application.
        """
        logger.debug("PROTECT: Postfilter for %s", self.name)

        evaluations_for_rule = self.evaluations_for_rule()

        context = contrast.CS__CONTEXT_TRACKER.current()
        response_code = context.response.status_code

        for evaluation in evaluations_for_rule:
            if str(response_code).startswith("4") or str(response_code).startswith("5"):
                attack = self.build_attack_without_match(
                    method=evaluation.value, response_code=response_code,
                )
            else:
                attack = self.build_attack_with_match(
                    None, method=evaluation.value, response_code=response_code,
                )
            self._append_to_activity(attack)

    def build_sample(self, evaluation, candidate_string, **kwargs):
        sample = self.build_base_sample(None)

        method = kwargs.get("method", "")

        sample.method_tampering.method = method
        sample.method_tampering.response_code = kwargs.get("response_code", -1)
        sample.user_input.CopyFrom(self.build_user_input(method))

        return sample

    def build_user_input(self, method):
        ui = UserInput()
        ui.key = self.USER_INPUT_KEY
        ui.input_type = UserInput.METHOD
        ui.value = method
        return ui



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.utils import get_properties
from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.utils.decorators import fail_safely

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class DBWritePropagator(BasePropagator):
    """
    Propagator to handle stored XSS. This propagator assumes the database
    column names are passed in via ALL_KWARGS.

    For each column name, patch the getter/setter property for the column name
    and create dynamic sources for these properties.
    """

    @property
    @fail_safely(
        "Failed to determine if DB_WRITE needs propagation", return_value=False
    )
    def needs_propagation(self):
        """
        Here, the source is ALL_KWARGS. self.sources[0] is the kwarg dictionary.
        Each key is a column name, and each value is the value for this particular
        ORM instance (corresponding to a DB row) in that column.

        We need to check if any of the values are tracked strings. If so, we save
        all relevant information to be used in propagate().
        """
        from contrast.agent.policy.applicator import build_method_name

        self.new_column_sources = {}
        cls = self.preshift.obj.__class__
        source = self.sources[0]

        for col_name, value in source.items():
            if not value:
                continue

            cs_method_name = build_method_name(col_name)
            if hasattr(cls, cs_method_name):
                continue

            col_value_properties = get_properties(value)
            if col_value_properties is None:
                continue

            self.new_column_sources[col_name] = col_value_properties

        return len(self.new_column_sources) > 0

    def track_target(self):
        """
        This propagator does not track the target because the target is
        an instance of an ORM model, such as django's Model class. If it
        were a string instead, we would track it.
        """
        pass

    def add_tags_and_properties(self, ret, frame):
        """
        Because the target is an instance of an ORM model,
        such as django's Model class, and not a string, we do this
        work inside self.propagate(). Later on we could refactor
        to move the work here.
        """
        pass

    def propagate(self):
        from contrast.agent.policy.applicator import (
            apply_patch_to_dynamic_property,
            save_original_method,
            build_method_name,
        )

        cls = self.preshift.obj.__class__

        for col_name, col_value_properties in self.new_column_sources.items():
            cs_method_name = build_method_name(col_name)
            old_property = getattr(cls, col_name)
            if not save_original_method(cls, cs_method_name, old_property):
                continue

            self.apply_tags(self.node, col_value_properties.origin)

            tags = [tag for tag in col_value_properties.tags]
            logger.debug("DB_WRITE adding column as dynamic source: %s", col_name)
            apply_patch_to_dynamic_property(cls, col_name, tags)

            col_value_properties.build_event(
                self.node,
                col_value_properties.origin,
                self.preshift.obj,
                self.target,
                self.preshift.args,
                self.preshift.kwargs,
                [],
                0,
                None,
            )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
Implementation of patches for the `re` module.

Other methods of the `re` module such as `split` and `escape` can be patched directly,
so they are implemented in policy.
"""
import functools
from contrast.extern import structlog as logging
import sys

from contrast.extern.wrapt import register_post_import_hook
from contrast.extern import six

import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.policy.loader import Policy
from contrast.agent.assess.apply_trigger import cs__apply_trigger
from contrast.agent.assess.policy.analysis import analyze
from contrast.agent.assess.policy.propagators import regex_propagator
from contrast.utils.decorators import fail_safely
from contrast.utils.patch_utils import patch_cls_or_instance

logger = logging.getLogger("contrast")

SPLIT = "split"
SUB = "sub"
SUBN = "subn"
MATCH = "match"
SEARCH = "search"
FINDALL = "findall"
FINDITER = "finditer"
FULLMATCH = "fullmatch"


def _group_propagator(method, propagator, *args, **kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    result = method(*args, **kwargs)
    if context is not None and context.propagate_assess:
        with scope.propagation_scope():
            propagator(args[0], result, *args[1:])
    return result


def group_hook(original_func, patch_policy, *args, **kwargs):
    propagator = regex_propagator.propagate_group
    return _group_propagator(original_func, propagator, *args, **kwargs)


def groups_hook(original_func, patch_policy, *args, **kwargs):
    propagator = regex_propagator.propagate_groups
    return _group_propagator(original_func, propagator, *args, **kwargs)


def groupdict_hook(original_func, patch_policy, *args, **kwargs):
    propagator = regex_propagator.propagate_groupdict
    return _group_propagator(original_func, propagator, *args, **kwargs)


def wrap_repl(repl):
    repl_results = []

    def new_repl(match):
        result = repl(match)
        repl_results.append(result)
        return result

    functools.update_wrapper(new_repl, repl)
    return new_repl, repl_results


@fail_safely("Failed to get propagation node")
def _get_propagation_node(name):
    policy = Policy().policy_by_name[name]
    return policy.propagator_nodes[0] if policy.propagator_nodes else None


@fail_safely("Failed to propagate sub(n)")
def _analyze_sub(node, retval, repl_results, args, kwargs, new_args):
    if node.method_name == "subn":
        result, count = retval
    else:
        # Account for the fact that count could either be positional or kwarg
        count = args[3] if len(args) == 4 else kwargs.pop("count", 0)
        result = retval

    # Omit count (and flags) if they are part of posargs since they are being passed
    # explicitly to our propagator
    new_args = new_args[:3]
    new_kwargs = dict(count=count)
    if not node.instance_method:
        # Account for the fact that flags could either be positional or kwarg
        new_kwargs["flags"] = args[4] if len(args) > 4 else kwargs.get("flags", 0)

    with scope.propagation_scope():
        regex_propagator.propagate_sub(
            node, result, repl_results, *new_args[:3], **new_kwargs
        )


def _sub_hook_impl(name, original_func, *args, **kwargs):
    """
    Hook for re.sub and re.subn used for propagation in assess

    The following explains why we can't simply patch these methods using
    policy.

    It is possible for the repl argument to be a callable. In this case, the
    callable is passed a Match object, and it returns the string to be used for
    the replacement. In order to correctly propagate the substitution
    operation, we need to keep track of the results of calling the replacement
    function.

    It might seem like we should just call the replacement function again
    during our propagation action. But this is not practicable for several
    reasons:

      1. We're in scope at the time, so any propagation that needs to occur
         within the replacement callable itself will be missed.
      2. Related to above, but methods of Match do not return the same object
         even when called multiple times with the same arguments, so we would
         not be tracking the strings that actually get used in the substitution
         result.
      3. There's no guarantee that the replacement function does not cause any
         side effects or rely on any state in application code. We definitely
         don't want to mess around with this.

    The solution is to wrap the replacement callable with our own function that
    records the results of each call. We then pass our wrapped callable to the
    original function, and we pass the accumulated results to the propagator.
    This has the additional benefit of allowing us to wrap the match object
    that is passed to the repl function with our proxied object so that we
    propagate any calls that are made within this function if necessary.
    """
    # Get the non-propagation case out of the way here
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None or not context.propagate_assess:
        return original_func(*args, **kwargs)

    node = _get_propagation_node(name)
    if not node:
        return original_func(*args, **kwargs)

    try:
        repl = args[1]
        new_repl, repl_results = wrap_repl(repl) if callable(repl) else (repl, None)
        new_args = tuple(args[:1]) + (new_repl,) + tuple(args[2:])
    except Exception:
        # This indicates that the original caller passed garbage, so call the original
        # function and let the error propagate back up to where they can clean up their
        # own mess.
        return original_func(*args, **kwargs)

    retval = original_func(*new_args, **kwargs)

    _trigger_redos(name, retval, args, kwargs)
    _analyze_sub(node, retval, repl_results, args, kwargs, new_args)

    return retval


@fail_safely("Failed to analyze redos trigger")
def _trigger_redos(name, result, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        return

    policy = Policy()
    rule = policy.triggers["redos"]

    if rule.disabled:
        # Given how often these patches are called within apps, it's well worth
        # returning early here even though this logic exists further down the stack.
        return

    if not scope.in_trigger_scope() and not scope.in_contrast_scope():
        trigger_node = _get_redos_trigger_node(name, rule)

        # Both trigger and contrast scope are needed here.
        with scope.trigger_scope(), scope.contrast_scope():
            # we cannot use trigger_policy.apply here due to the instance_method
            # logic it uses to remove self_obj from args. For the redos action,
            # specifically for re.Pattern.method nodes, we NEED the self_obj
            # to be in the args at this time.
            source = trigger_node.get_matching_sources(None, result, args, kwargs)[0]
            cs__apply_trigger(
                context, rule, trigger_node, source, None, result, None, args, kwargs,
            )


def _get_redos_trigger_node(trigger_name, rule):
    trigger_loc = "re"
    if "Pattern" in trigger_name:
        trigger_loc = "re.Pattern"

    func_name = trigger_name.split(".")[-1]
    trigger_nodes = rule.find_trigger_nodes(trigger_loc, func_name)

    return trigger_nodes[0]


def sub_hook(original_sub, patch_policy, *args, **kwargs):
    """See docstring for _sub_hook_impl above"""
    return _sub_hook_impl("re.sub", original_sub, *args, **kwargs)


def subn_hook(original_subn, patch_policy, *args, **kwargs):
    """See docstring for _sub_hook_impl above"""
    return _sub_hook_impl("re.subn", original_subn, *args, **kwargs)


def pattern_sub_hook(original_sub, patch_policy, *args, **kwargs):
    """See docstring for _sub_hook_impl above"""
    return _sub_hook_impl("re.Pattern.sub", original_sub, *args, **kwargs)


def pattern_subn_hook(original_subn, patch_policy, *args, **kwargs):
    """See docstring for _sub_hook_impl above"""
    return _sub_hook_impl("re.Pattern.subn", original_subn, *args, **kwargs)


def pattern_findall_hook(original_func, patch_policy, *args, **kwargs):
    retval = original_func(*args, **kwargs)
    _trigger_redos("re.Pattern.findall", retval, args, kwargs)
    _analyze_policy("re.Pattern.findall", retval, args, kwargs)
    return retval


@fail_safely("Failed to analyze")
def _analyze_policy(name, result, args, kwargs):
    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None or not context.propagate_assess:
        return

    with scope.contrast_scope():
        policy = Policy()
        patch_policy = policy.policy_by_name.get(name)
        if not patch_policy:
            return

    analyze(context, patch_policy, result, args, kwargs)


def pattern_split_hook(original_split, patch_policy, *args, **kwargs):
    result = original_split(*args, **kwargs)
    _trigger_redos("re.Pattern.split", result, args, kwargs)
    _analyze_policy("re.Pattern.split", result, args, kwargs)
    return result


def pattern_match_hook(original_func, patch_policy, *args, **kwargs):
    retval = original_func(*args, **kwargs)
    _trigger_redos("re.Pattern.match", retval, args, kwargs)
    return retval


def pattern_search_hook(original_func, patch_policy, *args, **kwargs):
    retval = original_func(*args, **kwargs)
    _trigger_redos("re.Pattern.search", retval, args, kwargs)
    return retval


def pattern_finditer_hook(original_func, patch_policy, *args, **kwargs):
    retval = original_func(*args, **kwargs)
    _trigger_redos("re.Pattern.finditer", retval, args, kwargs)
    return retval


def pattern_fullmatch_hook(original_func, patch_policy, *args, **kwargs):
    retval = original_func(*args, **kwargs)
    _trigger_redos("re.Pattern.fullmatch", retval, args, kwargs)
    return retval


def patch_re(re_module):
    # The re.Pattern and re.Match classes are not directly accessible in all versions
    # of Python, so we do this somewhat hacky workaround to get a reference to them.
    pattern_cls = re_module.compile("").__class__
    match_cls = re_module.match("", "").__class__

    patch_cls_or_instance(re_module, SUB, sub_hook)
    patch_cls_or_instance(re_module, SUBN, subn_hook)

    patch_cls_or_instance(pattern_cls, SUB, pattern_sub_hook)
    patch_cls_or_instance(pattern_cls, SUBN, pattern_subn_hook)
    patch_cls_or_instance(pattern_cls, SPLIT, pattern_split_hook)
    patch_cls_or_instance(pattern_cls, FINDALL, pattern_findall_hook)

    patch_cls_or_instance(pattern_cls, MATCH, pattern_match_hook)
    patch_cls_or_instance(pattern_cls, SEARCH, pattern_search_hook)
    patch_cls_or_instance(pattern_cls, FINDITER, pattern_finditer_hook)

    if six.PY3:
        # fullmatch method is new in Py3
        patch_cls_or_instance(pattern_cls, FULLMATCH, pattern_fullmatch_hook)

    patch_cls_or_instance(match_cls, "group", group_hook)
    patch_cls_or_instance(match_cls, "groups", groups_hook)
    patch_cls_or_instance(match_cls, "groupdict", groupdict_hook)


def register_patches():
    register_post_import_hook(patch_re, "re")


def reverse_patches():
    re_module = sys.modules.get("re")
    if not re_module:
        return

    pattern_cls = re_module.compile("").__class__
    match_cls = re_module.match("", "").__class__

    patch_manager.reverse_patches_by_owner(re_module)
    patch_manager.reverse_patches_by_owner(pattern_cls)
    patch_manager.reverse_patches_by_owner(match_cls)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import re

from contrast.agent.protect.rule.base_rule import BaseRule
from contrast.agent.protect.rule.xxe.entity_wrapper import EntityWrapper
from contrast.api.dtm_pb2 import RaspRuleSample, UserInput, XxeMatch, XxeWrapper
from contrast.utils.stack_trace_utils import StackTraceUtils


class Xxe(BaseRule):
    """
    XXE Protection rule
    """

    NAME = "xxe"
    INPUT_NAME = "XML Prolog"

    EXTERNAL_ENTITY_PATTERN = re.compile(
        r"<!ENTITY\s+[a-zA-Z0-f]+\s+(?:SYSTEM|PUBLIC)\s+(.*?)>"
    )

    def __init__(self, settings):
        BaseRule.__init__(self, settings)

        self.prolog_xml = None

    def is_prefilter(self):
        return False

    def is_postfilter(self):
        return False

    def find_attack(self, candidate_string=None, **kwargs):
        if self.protect_excluded_by_code():
            return None

        last_idx = 0
        declared_entities = []
        entities_resolved = []

        for match in self.EXTERNAL_ENTITY_PATTERN.finditer(candidate_string):
            last_idx = match.end(0)

            entity_wrapper = EntityWrapper(match.group())
            if not entity_wrapper.is_external_entity():
                continue

            declared_entities.extend([self._build_match(match.group(), last_idx)])
            entities_resolved.extend([self._build_wrapper(entity_wrapper)])

        self.prolog_xml = candidate_string[:last_idx]

        attack = self.build_attack_with_match(
            candidate_string,
            declared_entities=declared_entities,
            entities_resolved=entities_resolved,
        )
        return attack

    def _build_match(self, match_string, end_idx):
        match = XxeMatch()
        match.end_idx = end_idx
        match.start_idx = end_idx - len(match_string)
        return match

    def _build_wrapper(self, entity_wrapper):
        wrapper = XxeWrapper()
        wrapper.system_id = entity_wrapper.system_id
        wrapper.public_id = entity_wrapper.public_id
        return wrapper

    def build_sample(self, evaluation, url, **kwargs):
        sample = self.build_base_sample(evaluation)
        declared_entities = kwargs["declared_entities"]
        if declared_entities:
            sample.xxe.declared_entities.extend(declared_entities)

        entities_resolved = kwargs["entities_resolved"]
        if entities_resolved:
            sample.xxe.entities_resolved.extend(entities_resolved)

        sample.xxe.xml = self.prolog_xml
        return sample

    def build_base_sample(self, evaluation):
        sample = RaspRuleSample()
        sample.stack_trace_elements.extend(StackTraceUtils.build(ignore=True))

        sample.user_input.CopyFrom(self.build_user_input(evaluation))

        return sample

    def build_user_input(self, evaluation):
        ui = UserInput()
        ui.key = self.INPUT_NAME
        ui.input_type = UserInput.UNKNOWN
        ui.value = self.prolog_xml
        return ui

    def infilter_kwargs(self, user_input, patch_policy):
        return dict(framework=patch_policy.method_name)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.webob.headers import ResponseHeaders
from contrast.agent.middlewares.response_wrappers.base_response_wrapper import (
    BaseResponseWrapper,
)


class DjangoResponseWrapper(BaseResponseWrapper):
    def __init__(self, response):
        self._response = response
        self._streaming_cache = None

    @property
    def body(self):
        if self._response.streaming:
            return self._get_streaming_content()
        return self._response.content

    @property
    def headers(self):
        """
        In django, response headers are normalized. The outermost key is the normalized header
        value, and value is a tuple of (original_header_key, header_value). For example:
        {'content-type': ('CoNtEnt-TyPe', 'text/html'), .... }

        Conveniently, we can pass these values into the constructor for webob's ResponseHeaders
        object.

        :return: webob.headers.ResponseHeaders object which conforms to BaseResponseWrapper's
            requirements for this property.
        """
        return ResponseHeaders(self._response._headers.values())

    @property
    def status_code(self):
        return self._response.status_code

    def _get_streaming_content(self):
        """
        Safely extract the content of a streaming response body. This method guarantees that the
        response body is restored after extraction. Unfortunately, this is likely a performance hit
        to applications streaming large response bodies.

        :return: body of a streaming response as bytes
        """
        if not self._streaming_cache:
            body = b"".join(self._response.streaming_content)
            self._response.streaming_content = [body]
            self._streaming_cache = body
            return self._streaming_cache

        return self._streaming_cache



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")

from contrast.agent import scope
from contrast.agent.policy.applicator import register_import_hooks
from contrast.agent.settings_state import SettingsState
from contrast.assess_extensions import cs_str
from contrast.patches import (
    cs_str as cs_str_patches,
    register_assess_patches,
    register_common_patches,
    register_library_patches,
)
from contrast.utils.namespace import Namespace
from contrast.utils.patch_utils import repatch_imported_modules


class module(Namespace):
    hook = None
    enabled = False


def enable_assess_patches():
    """
    Enables extension hooks and other string patches.

    Has no effect if these patches are already enabled.
    """
    if module.enabled:
        return

    try:
        module.hook = cs_str.enable(logger)
    except RuntimeError:
        logger.error(
            "Local python builds on OSX may lead to 'Failed to unprotect memory'"
        )
        logger.error(
            "If this is applies to you, try running `contrast-fix-interpreter-permissions`"
        )
        raise

    # Enable any string patches that need to be applied through Python
    cs_str_patches.patch_strtype_methods()

    module.enabled = True


def disable_assess_patches():
    """
    Disables extension hooks and other string patches.

    Has no effect if these patches are not already enabled.
    """
    if not module.enabled:
        return

    cs_str.disable(module.hook)

    # Disable any string patches that were applied through Python
    cs_str_patches.unpatch_strtype_methods()

    module.enabled = False


def _enable_patches():
    settings = SettingsState()

    if settings.is_analyze_libs_enabled():
        register_library_patches()

    if settings.is_protect_enabled():
        register_common_patches()

        logger.debug("adding protect policy")
        register_import_hooks(protect_mode=True)

        # This has no effect if the patches are not enabled
        disable_assess_patches()

    if settings.is_assess_enabled():
        enable_assess_patches()

        logger.debug("enabled assess string patches")
        register_common_patches()
        register_assess_patches()

        logger.debug("adding assess policy")
        register_import_hooks()

    logger.debug("revisiting imported modules to apply patches")
    repatch_imported_modules()


def enable_patches():
    """
    Enable all patches for agent based on current settings
    """
    # Being in scope here prevents us from inadvertently propagating while we are
    # applying patches and navigating policy. This has a fairly significant performance
    # impact for assess initialization, and also promotes correctness/safety.
    with scope.contrast_scope():
        _enable_patches()



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
"""
This module contains workarounds for the fact that some builtin methods appear
to be unpatchable using funchook for one reason or another.
"""
import sys

from contrast.extern.six import PY2, PY3, binary_type, text_type

import contrast
from contrast.agent import scope
from contrast.agent.policy import patch_manager
from contrast.agent.assess.policy.analysis import skip_analysis
from contrast.agent.policy.loader import Policy
from contrast.agent.assess.policy.propagation_policy import (
    PROPAGATOR_ACTIONS,
    propagate,
)
from contrast.agent.assess.policy.preshift import Preshift
from contrast.assess_extensions import smart_setattr
from contrast.utils.patch_utils import patch_cls_or_instance


from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


PY38 = PY3 and sys.version_info[1] == 8


def bytearray_join_patch(orig_method, patch_policy, self, *args, **kwargs):
    # Since we need to make reference to the input multiple times, convert the
    # first argument to a list and use that instead. This prevents any iterators
    # from being exhausted before we can make use of them in propagation.
    # For bytearray.join, args == (list_or_iterator_of_things_to_join,...)
    # Note that this is different from the C hooks for other join methods. In
    # those cases, the PyObject *args argument corresponds to just the list or
    # iterator itself, in contrast to a tuple that contains that list or
    # iterator. (Got that straight?)
    args_list = [list(args[0])] + list(args[1:])
    result = orig_method(self, *args_list, **kwargs)

    context = contrast.CS__CONTEXT_TRACKER.current()
    if skip_analysis(context) or scope.in_scope():
        return result

    try:
        frame = sys._getframe()
        with scope.propagation_scope():
            propagate("join", result, self, result, frame, args_list, kwargs)
    except Exception:
        logger.exception("failed to propagate bytearray.join")
    finally:
        return result


def strtype_join_patch(orig_method, patch_policy, self, *args, **kwargs):
    args_list = list(args[0])

    with scope.propagation_scope():
        result = orig_method(self, args_list, **kwargs)

    context = contrast.CS__CONTEXT_TRACKER.current()
    if skip_analysis(context) or scope.in_scope():
        return result

    try:
        frame = sys._getframe()
        with scope.propagation_scope():
            propagate("join", result, self, result, frame, (args_list,), kwargs)
    except Exception:
        with scope.propagation_scope():
            logger.exception("failed to propagate join")
    finally:
        return result


def str_format_patch(orig_method, patch_policy, self, *args, **kwargs):
    """
    Propagation hook for str.format

    This hook is a special case because we need to enable some propagation to occur
    while we evaluate whether to propagate this particular event. With the current
    general hook infrastructure, this is not possible, so we need to account for it
    here. Eventually it may be possible to fit this back into the more general
    infrastructure if we overhaul the way that scope works.
    """
    result = orig_method(self, *args, **kwargs)

    context = contrast.CS__CONTEXT_TRACKER.current()
    if skip_analysis(context) or scope.in_scope():
        return result

    try:
        with scope.contrast_scope():
            frame = sys._getframe()

            policy = Policy()
            preshift = Preshift(self, args, kwargs)

        for node in policy.propagators_by_name["BUILTIN.str.format"]:
            propagator_class = PROPAGATOR_ACTIONS.get(node.action)
            if propagator_class is None:
                continue

            propagator = propagator_class(node, preshift, result)

            # This evaluation must not occur in scope. This is what enables us
            # to perform any conversions from object to __str__ or __repr__,
            # while allowing propagation to occur through those methods if
            # necessary.
            if not propagator.needs_propagation:
                continue

            with scope.contrast_scope():
                propagator.track_and_propagate(result, frame)
    except Exception:
        logger.exception("failed to propagate str.format")
    finally:
        return result


def generic_str_patch(orig_method, patch_policy, self, *args, **kwargs):
    """
    General-purpose patch implementation to be used for replacing string methods
    """
    result = orig_method(self, *args, **kwargs)

    context = contrast.CS__CONTEXT_TRACKER.current()
    if skip_analysis(context) or scope.in_scope():
        return result

    try:
        frame = sys._getframe()
        with scope.propagation_scope():
            propagate(orig_method.__name__, result, self, result, frame, args, kwargs)
    except Exception:
        name = orig_method.__class__.__name__
        logger.exception("failed to propagate %s.%s", name, orig_method.__name__)
    finally:
        return result


# TODO: PYT-1452 eventually these patches should only be applied on Py38 on Alpine
alpine_py38_buggy_methods = [
    "capitalize",
    "casefold",
    "title",
    "lower",
    "upper",
    "swapcase",
]


def patch_py38_methods():
    """
    Apply patches for methods that can't be hooked with funchook in Py38

    Specifically, the `strip`, `lstrip`, and `rstrip` methods of both `str` and
    `bytes` can't be hooked with funchook.
    """
    for strtype in [str, bytes]:
        # TODO: PYT-1452 eventually these patches should only be applied on Py38 on Alpine
        for method_name in ["strip", "lstrip", "rstrip"] + alpine_py38_buggy_methods:
            if hasattr(strtype, method_name):
                patch_cls_or_instance(strtype, method_name, generic_str_patch)


def patch_format_methods():
    patch_cls_or_instance(str, "format", str_format_patch)

    if PY2:
        patch_cls_or_instance(text_type, "format", str_format_patch)


def property_getter(self):
    return contrast.STRING_TRACKER.get(self, None)


def property_setter(self, value):
    contrast.STRING_TRACKER.update_properties(self, value)


def enable_str_properties():
    strprop = property(fget=property_getter, fset=property_setter)

    smart_setattr(text_type, "cs__properties", strprop)
    smart_setattr(binary_type, "cs__properties", strprop)
    smart_setattr(bytearray, "cs__properties", strprop)


def patch_strtype_methods():
    """
    Apply patches for all methods that can't be hooked with funchook

    For reasons that are not well understood, funchook fails to apply patches
    to certain methods on certain platforms. Now that we can hook builtins
    directly from Python, we apply the workarounds here. There is probably a
    performance penalty for having these patches implemented purely in Python,
    but it's better than not hooking these methods at all.

    Currently the only versions affected are Py3 and higher, but if we ever
    have problems with Py2 hooks, those workarounds should be added here too.
    """
    enable_str_properties()

    patch_cls_or_instance(text_type, "partition", generic_str_patch)
    patch_cls_or_instance(text_type, "rpartition", generic_str_patch)

    patch_format_methods()

    # Patching bytearray.join with funchook does not work for any Py3 versions
    if PY3:
        patch_cls_or_instance(bytearray, "join", bytearray_join_patch)

    if PY38:
        patch_py38_methods()
        # TODO: PYT-1452 eventually these patches should be applied only on Alpine
        patch_cls_or_instance(text_type, "join", strtype_join_patch)
        patch_cls_or_instance(binary_type, "join", strtype_join_patch)


def unpatch_strtype_methods():
    """
    Replace all patched strtype methods with the original implementation
    """
    patch_manager.reverse_patches_by_owner(text_type)
    patch_manager.reverse_patches_by_owner(binary_type)
    patch_manager.reverse_patches_by_owner(bytearray)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.agent.assess.policy.propagators.base_propagator import BasePropagator
from contrast.agent.assess.utils import copy_from
from contrast.agent.assess.utils import get_properties


class KeepPropagator(BasePropagator):
    def propagate(self):
        target_props = get_properties(self.target)
        if target_props is None:
            return

        self.first_source = self.sources[0]
        copy_from(self.target, self.first_source, 0, self.node.untags)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from contrast.extern.wrapt import register_post_import_hook

from contrast.utils.patch_utils import patch_cls_or_instance

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


PYCRYPTO = "pycrypto"
DECRYPT = "decrypt"


def __cs_decrypt(original_decrypt, *args, **kwargs):
    """
    :param original_decrypt: decrypt_method
    :param args: 1 => self, 2 => cipher_text
    :return:

    If the mode is CBC attempt to decrypt the cipher_text and validate the padding.
    If the padding is invalid, apply the padding oracle rule.
    """
    from Crypto.Cipher.blockalgo import MODE_CBC

    if args[0].mode == MODE_CBC:
        value = original_decrypt(*args, **kwargs)

        try:
            if not __cs_validate_padding(value.decode()):
                # the padding oracle rule is (temporarily?) removed from the agent
                # apply_rule("Invalid Padding")
                pass
        except UnicodeDecodeError:
            _announce_unicode_decode_error()

        return value

    return original_decrypt(*args, **kwargs)


def __cs_validate_padding(padded_text):
    return all([n == padded_text[-1] for n in padded_text[-ord(padded_text[-1]) :]])


def _announce_unicode_decode_error():
    logger.warning(
        "Crypto decryption could not be decoded with UTF-8. PaddingOracleRule not applied."
    )


def patch_pycrypto(crypto_module):
    patch_cls_or_instance(
        crypto_module.Cipher.blockalgo.BlockAlgo, DECRYPT, __cs_decrypt
    )


def register_patches():
    register_post_import_hook(patch_pycrypto, "Crypto")



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import mock
import django

import pytest
from contrast.extern import six

from contrast.agent.policy import constants
from contrast.agent.assess.policy import source_policy
from contrast.agent.middlewares.base_middleware import BaseMiddleware
from contrast.api.dtm_pb2 import TraceEvent
from contrast.api.settings_pb2 import AgentSettings
from contrast.test.settings_builder import SettingsBuilder
from contrast.test.contract.events import check_event
from contrast.utils import patch_utils


@pytest.fixture(autouse=True)
def reset_singletons():
    """
    We delete the existing Singleton instances before every test to prevent a test from
    inadvertently using the same object.
    """
    # Import here to prevent circular imports.
    from contrast.agent import service_client
    from contrast.agent.settings_state import SettingsState

    SettingsState.clear_instance()
    service_client._ServiceState.clear_instance()
    yield


@pytest.fixture(autouse=True)
def disable_service_client(mocker):
    """Prevent the test suite from inadvertently starting the service"""
    mocker.patch("contrast.agent.service_client.time.sleep")
    mocker.patch("contrast.agent.service_client._ServiceState.start_service")
    mocker.patch(
        "contrast.agent.service_client._ServiceState.send_dtm",
        return_value=AgentSettings(),
    )


@pytest.fixture(autouse=True)
def ensure_no_scope():
    """
    Before each test assert that agent is not in scope
    to ensure prevent unexpected behavior due to scope leaking.

    If any of these assertions fail, check the tests that ran before the failure to see
    where the scope needs to be exited.
    """
    from contrast.assess_extensions.cs_str import get_thread_scope

    assert get_thread_scope() == (0, 0, 0, 0)


@pytest.fixture(autouse=True)
def modified_modules_to_ignore_prefixes(monkeypatch):
    """
    In most uses of the agent we ignore module such as contrast...
    so we don't bother patching them, but since our framework tests often start
    with contrast., we need to make sure these modules are not in the
    modules to ignore.
    """
    new_prefixes = list(patch_utils.MODULES_TO_IGNORE_PREFIXES[:])
    new_prefixes.remove("contrast.")
    new_prefixes = tuple(new_prefixes)

    monkeypatch.setattr(patch_utils, "MODULES_TO_IGNORE_PREFIXES", new_prefixes)


@pytest.fixture(autouse=True)
def reset_base_middleware():
    BaseMiddleware._loaded = False
    yield


@pytest.fixture(autouse=True, scope="module")
def disable_heartbeat():
    patch_heartbeat = mock.patch(
        "contrast.agent.settings_state.SettingsState.establish_heartbeat"
    )

    patch_heartbeat.start()
    yield
    patch_heartbeat.stop()


@pytest.fixture(autouse=True, scope="module")
def disable_providers():
    """
    Providers iterate over all loaded modules so it's an expensive
    """
    patch_enable_providers = mock.patch(
        "contrast.agent.middlewares.base_middleware.enable_providers"
    )

    patch_enable_providers.start()

    yield

    patch_enable_providers.stop()


def assert_django_sqli_finding_events(finding, source_class_name):
    assert len(finding.events) == 5

    check_event(
        finding.events[0],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(constants.CREATION_TYPE),
        class_name=source_class_name,
        method_name="__getitem__",
        source_types=[source_policy.PARAMETER_TYPE],
        first_parent=None,
    )

    check_event(
        finding.events[1],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.PARAMETER + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="django.utils.safestring",
        method_name="mark_safe",
        source_types=[],
        first_parent=finding.events[0],
    )

    check_event(
        finding.events[2],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="concat",
        source_types=[],
        first_parent=finding.events[1],
    )

    check_event(
        finding.events[3],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="concat",
        source_types=[],
        first_parent=finding.events[2],
    )

    check_event(
        finding.events[4],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(constants.TRIGGER_TYPE),
        class_name="SQLite3",
        method_name="execute",
        source_types=[],
        first_parent=finding.events[3],
    )


def assert_xss_cookie_finding_events(finding):
    legacy_django = django.__version__ < "1.11"

    assert len(finding.events) == 9

    check_event(
        finding.events[0],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(constants.CREATION_TYPE),
        class_name="django.http.request.HttpRequest",
        method_name="__getattribute__",
        source_types=[source_policy.COOKIE_TYPE],
        first_parent=None,
    )

    check_event(
        finding.events[1],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="CAST",
        source_types=[],
        first_parent=finding.events[0],
    )

    check_event(
        finding.events[2],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="cformat",
        source_types=[],
        first_parent=finding.events[1],
    )

    check_event(
        finding.events[3],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="join",
        source_types=[],
        first_parent=finding.events[2],
    )

    check_event(
        finding.events[8],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(constants.TRIGGER_TYPE),
        class_name="django.core.handlers.base" if not legacy_django else "app",
        method_name="_get_response" if not legacy_django else "app",
        source_types=[],
        first_parent=finding.events[7],
    )


def assert_subprocess_popen_finding_events(findings):
    """
    Verify each event in the finding. This requires some finagling because each
    combination of django version + python version has a slightly different
    sequence of events.
    """
    extra_kwargs = {}
    if six.PY2:

        def taint_validator(taint_ranges, *args):
            assert taint_ranges[0].range == taint_ranges[2].range == "0:4"
            assert taint_ranges[1].range == taint_ranges[3].range == "5:11"

        extra_kwargs["taint_range_validator"] = taint_validator

    events = findings[0].events
    assert len(events) == 10 if django.VERSION[:2] == (1, 9) and six.PY3 else 9
    check_event(
        events[0],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(constants.CREATION_TYPE),
        class_name="wsgi.environ",
        method_name="QUERY_STRING",
        source_types=[source_policy.QUERYSTRING_TYPE],
        first_parent=None,
        source="",
        target=constants.RETURN,
        ret_value="user_input=echo+attack",
    )

    offset = 1 if six.PY2 else 0
    offset += 1 if django.VERSION[:2] == (1, 9) and six.PY2 else 0
    check_event(
        events[1 + offset],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.OBJECT_KEY + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="encode" if six.PY3 else "split",
        source_types=[],
        first_parent=events[0 + offset],
        source=constants.OBJECT,
        target=constants.RETURN,
        ret_value="user_input=echo+attack" if six.PY3 else "echo+attack",
    )

    check_event(
        events[2 + offset],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.OBJECT_KEY + constants.TO_MARKER + constants.RETURN_KEY
        )
        if six.PY3
        else TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="str",
        method_name="decode" if six.PY3 else "replace",
        source_types=[],
        first_parent=events[1 + offset],
        source=constants.OBJECT if six.PY3 else "OBJ,P1",
        target=constants.RETURN,
        ret_value="user_input=echo+attack" if six.PY3 else "echo attack",
        **extra_kwargs
    )

    offset += 1 if six.PY3 else 0

    if django.VERSION[:2] == (1, 9):
        check_event(
            events[3 + offset],
            event_type=TraceEvent.TYPE_PROPAGATION,
            action=TraceEvent.Action.Value(
                constants.OBJECT_KEY + constants.TO_MARKER + constants.RETURN_KEY
            ),
            class_name="str",
            method_name="split",
            source_types=[],
            first_parent=events[2 + offset],
            source=constants.OBJECT,
            target=constants.RETURN,
            ret_value="user_input=echo+attack" if six.PY3 else "echo attack",
            **extra_kwargs
        )
    else:
        check_event(
            events[3 + offset],
            event_type=TraceEvent.TYPE_PROPAGATION,
            action=TraceEvent.Action.Value(
                constants.OBJECT_KEY + constants.TO_MARKER + constants.RETURN_KEY
            ),
            class_name="str",
            method_name="split",
            source_types=[],
            first_parent=events[2 + offset],
            source=constants.OBJECT,
            target=constants.RETURN,
            ret_value="echo+attack" if six.PY3 else "echo attack",
            **extra_kwargs
        )
    if django.VERSION[:2] == (1, 9):
        offset += 1 if six.PY3 else -1
    else:
        check_event(
            events[4 + offset],
            event_type=TraceEvent.TYPE_PROPAGATION,
            action=TraceEvent.Action.Value(
                constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
            ),
            class_name="str" if six.PY3 else "urllib",
            method_name="replace" if six.PY3 else "unquote",
            source_types=[],
            first_parent=events[3 + offset],
            source="OBJ,P1" if six.PY3 else "P0,KWARG:s",
            target=constants.RETURN,
            ret_value="echo attack",
            **extra_kwargs
        )

    check_event(
        events[5 + offset],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(
            constants.ALL_TYPE + constants.TO_MARKER + constants.RETURN_KEY
        )
        if six.PY3
        else TraceEvent.Action.Value(
            constants.OBJECT_KEY + constants.TO_MARKER + constants.RETURN_KEY
        ),
        class_name="urllib.parse" if six.PY3 else "str",
        method_name="unquote" if six.PY3 else "decode",
        source_types=[],
        first_parent=events[4 + offset],
        source="P0,KWARG:string" if six.PY3 else constants.OBJECT,
        target=constants.RETURN,
        ret_value="echo attack",
    )

    check_event(
        events[6 + offset],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(constants.CREATION_TYPE),
        class_name="django.utils.datastructures.MultiValueDict",
        method_name="__getitem__",
        source_types=[source_policy.PARAMETER_TYPE],
        first_parent=None,  # TODO: PYT-1464 simplify traces with multiple sources
        source="",
        target=constants.RETURN,
        ret_value="echo attack",
    )

    check_event(
        events[7 + offset],
        event_type=TraceEvent.TYPE_PROPAGATION,
        action=TraceEvent.Action.Value(constants.TRIGGER_TYPE),
        class_name="subprocess.Popen",
        method_name="__init__",
        source_types=[],
        first_parent=events[6 + offset],
        source="P0,KWARG:args",
        target="",
        ret_value="None",
        # No taint range expected on return value from Popen.__init__
        taint_range_validator=None,
    )


def validate_observed_route(request_context, method, url, source_name, source_type):
    observed_route = request_context.observed_route

    assert observed_route.verb == method
    assert observed_route.url == url

    sources = observed_route.sources
    matching_source = [src for src in sources if src.name == source_name][0]

    assert matching_source.name == source_name
    assert matching_source.type == source_type


def assert_route_appended(findings, url, verb):
    """
    Asserts that 1 route with the url and verb are in the findings

    Fail if we never found a matching route
    """
    for finding in findings:
        assert finding.version == constants.CURRENT_FINDING_VERSION
        for route in finding.routes:
            if route.url == url and route.verb == verb:
                return

    routes_message = " Routes: {}".format(findings[0].routes) if findings else ""
    pytest.fail("Unable to find route in findings;" + routes_message)


def build_assess_settings():
    return SettingsBuilder().build_assess()


def find_routes_by_verb(routes, verb):
    return [x for x in routes if x.verb == verb]


def current_routes_by_url(routes, url):
    return [x for x in routes.values() if x.url == url]


def assert_routes_count(routes, routes_count):
    assert len(routes) == routes_count



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent import scope
from contrast.agent.assess.policy.preshift import Preshift
from contrast.agent.assess.policy import propagation_policy
from contrast.agent.assess.policy import source_policy
from contrast.agent.assess.policy import trigger_policy
from contrast.agent.assess.utils import get_self_for_method
from contrast.agent.settings_state import SettingsState
from contrast.utils.decorators import fail_safely


@fail_safely("Failed to perform assess analysis.")
def analyze(context, patch_policy, result, args, kwargs):
    if skip_analysis(context):
        return

    with scope.contrast_scope():
        self_obj = get_self_for_method(patch_policy, args)
        preshift = Preshift(self_obj, args, kwargs)

        _analyze(patch_policy, preshift, self_obj, result, args, kwargs)


def skip_analysis(context):
    """
    Skip analysis if there is no context, scope, or configuration is False
    :param context: RequestContext
    :return:
    """
    if not context:
        return True
    if scope.in_contrast_scope():
        return True
    return not SettingsState().is_assess_enabled()


def _analyze(patch_policy, preshift, self_obj, ret, args, kwargs=None):
    if not patch_policy:
        return

    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None:
        return

    if patch_policy.trigger_nodes:
        # Each node may potentially correspond to a different rule
        for node in patch_policy.trigger_nodes:
            if not scope.in_trigger_scope():
                with scope.trigger_scope():
                    trigger_policy.apply(node.rule, [node], ret, args, kwargs)

    if patch_policy.source_nodes:
        source_policy.apply(patch_policy.source_nodes, self_obj, ret, args, kwargs)

    if not scope.in_propagation_scope():
        if patch_policy.propagator_nodes:
            with scope.propagation_scope():
                propagation_policy.apply(patch_policy.propagator_nodes, preshift, ret)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import contrast
from contrast.agent.assess.apply_trigger import build_finding
from contrast.agent.policy.loader import Policy
from contrast.agent.assess.policy.analysis import skip_analysis


def apply_rule(string, ptr_val, frame):
    """
    Applies unsafe-code-execution rule for exec statement in PY2

    This function is only ever called from extension code. In order to
    determine whether the rule has been violated, we need to do a reverse
    lookup based on the pointer value of the underlying char buffer that
    corresponds to the string object. This is because our hook for the exec
    statement does not see a PyObject but instead is only passed the underlying
    char buffer.
    """
    props = contrast.STRING_TRACKER.lookup_by_pointer(ptr_val)
    if props is None:
        return

    # TODO: PYT-915 handle the unicode case where we need to remove the propagation
    # event that is created when the string is encoded into UTF-8. Since this
    # doesn't happen in user-controlled or library code, it will be confusing
    # when reporting, so we should try to remove it. Ideally we will be able to
    # just grab the properties belonging to the pre-encoded unicode object.

    context = contrast.CS__CONTEXT_TRACKER.current()
    if context is None or skip_analysis(context):
        return

    policy = Policy()
    trigger_rule = policy.triggers["unsafe-code-execution"]
    trigger_nodes = trigger_rule.find_trigger_nodes("BUILTIN", "exec")

    for node in trigger_nodes:
        if trigger_rule.is_violated_properties(node, props):
            args = (string,)
            kwargs = {}

            build_finding(
                context,
                trigger_rule,
                node,
                string,
                None,
                None,
                None,
                args,
                kwargs,
                target_properties=props,
                frame=frame,
            )



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os

import pytest

from contrast.test.framework_test.base_django_test import BaseDjangoTest
from contrast.test.framework_test_utils import assert_subprocess_popen_finding_events
from contrast.test.contract.findings import (
    validate_finding_in_context,
    validate_ssrf_finding_no_mock,
    remove_redos_findings,
)
from contrast.test.helper import python2_only
from contrast.test.helper import python3_only


class VulnpyAssessTestMixin(object):
    @pytest.mark.xfail(reason="TODO: PYT-1235 No finding reported for cmdi")
    @python3_only
    def test_non_utf8_decoding_cmdi(self):
        """
        param_val a non utf8 encoded string. This is the result of shift_jis encoding of
        '手袋'. This encoding is automatically done by chrome, not in our test client. So
        we manually provide the encoded bytes here.
        """
        trigger = "os-system"
        rule_id = "cmd-injection"
        # Couldn't rely on passing 手袋 in the request because the test client doesn't
        # use shift_jis to encode.
        param_val = b"\x8e\xe8\x91\xdc"
        # Since we fail to decode param_value using utf-8 we rely on replace error
        # handling to get this new string
        expected_value = "���"
        path = "/vulnpy/cmdi/{}".format(trigger)

        self.client.get(path, {"user_input": param_val})

        validate_finding_in_context(self.request_context, rule_id, expected_value)

    @pytest.mark.parametrize("trigger", ["os-system", "subprocess-popen"])
    def test_cmdi(self, trigger):
        user_input = "echo attack"
        self.client.get("/vulnpy/cmdi/{}".format(trigger), {"user_input": user_input})

        validate_finding_in_context(self.request_context, "cmd-injection", user_input)

        if (
            isinstance(self, BaseDjangoTest)
            and trigger == "subprocess-popen"
            and not os.environ.get("USE_WSGI")  # TODO: PYT-1102
        ):
            assert_subprocess_popen_finding_events(
                self.request_context.activity.findings
            )

    @pytest.mark.parametrize(
        "trigger", ["pickle-load", "pickle-loads", "yaml-load", "yaml-load-all"]
    )
    def test_untrusted_deserialization(self, trigger):
        user_input = "some data to deserialize"
        try:
            self.client.get(
                "/vulnpy/deserialization/{}".format(trigger), {"user_input": user_input}
            )
        except Exception:
            # this view function doesn't handle its own errors, so we do it here
            pass

        validate_finding_in_context(
            self.request_context, "untrusted-deserialization", user_input
        )

    @pytest.mark.parametrize("trigger", ["io-open", "open"])
    def test_path_traversal(self, trigger):
        user_input = "does not matter"
        self.client.get("/vulnpy/pt/{}".format(trigger), {"user_input": user_input})

        validate_finding_in_context(self.request_context, "path-traversal", user_input)

    @python2_only
    def test_path_traversal_execfile(self):
        user_input = "does not matter"
        self.client.get("/vulnpy/pt/execfile", {"user_input": user_input})

        validate_finding_in_context(self.request_context, "path-traversal", user_input)

    @pytest.mark.parametrize(
        "trigger", ["sqlite3-execute", "sqlite3-executemany", "sqlite3-executescript"]
    )
    def test_sqli(self, trigger):
        user_input = "any value works"
        self.client.get(
            "/vulnpy/sqli/{}".format(trigger), params={"user_input": user_input}
        )

        # 3 queries per vulnpy request: database reset, attack, fetch results
        assert self.request_context.activity.query_count == 3
        validate_finding_in_context(self.request_context, "sql-injection", user_input)

    @pytest.mark.parametrize(
        "trigger", ["legacy-urlopen", "urlopen-str", "urlopen-obj"]
    )
    @pytest.mark.parametrize("safe", [False, True])
    def test_ssrf_urllib(self, trigger, safe):
        user_input = "not.a.url" if safe else "http://attacker.com/?q=foobar"
        self.client.get("/vulnpy/ssrf/{}".format(trigger), {"user_input": user_input})

        if not safe:
            validate_ssrf_finding_no_mock(trigger, self.request_context)
        else:
            findings = remove_redos_findings(
                self.request_context.activity.findings, "ssrf"
            )
            assert len(findings) == 0

    @pytest.mark.parametrize("trigger_class", ["httpconnection", "httpsconnection"])
    @pytest.mark.parametrize(
        "trigger_method", ["request-method", "putrequest-method", "init"]
    )
    def test_ssrf_httplib(self, trigger_method, trigger_class):
        trigger = "-".join([trigger_class, trigger_method])
        user_input = "www.attacker.com" if trigger_method == "init" else "DELETE"
        self.client.get("/vulnpy/ssrf/{}".format(trigger), {"user_input": user_input})

        validate_ssrf_finding_no_mock(trigger, self.request_context)

    @pytest.mark.parametrize("trigger_class", ["httpconnection", "httpsconnection"])
    @pytest.mark.parametrize("trigger_method", ["request", "putrequest"])
    def test_ssrf_httplib_safe(self, trigger_method, trigger_class):
        """
        HTTP*Connection.*request isn't vulnerable to SSRF via the "url" argument.
        This is because it only takes a path / querystring, which isn't vulnerable
        to ssrf.
        """
        trigger = "-".join([trigger_class, trigger_method, "url"])
        self.client.get("/vulnpy/ssrf/{}".format(trigger), {"user_input": "/some/path"})

        findings = remove_redos_findings(self.request_context.activity.findings, "ssrf")

        assert len(findings) == 0

    @pytest.mark.parametrize("trigger", ["random", "randint", "randrange", "uniform"])
    def test_weak_randomness(self, trigger):
        user_input = "abc123"
        self.client.get("/vulnpy/rand/{}".format(trigger), {"user_input": user_input})

        validate_finding_in_context(
            self.request_context, "crypto-weak-randomness", user_input, nondataflow=True
        )

    @pytest.mark.parametrize("trigger", ["hashlib-md5", "hashlib-sha1", "hashlib-new"])
    def test_insecure_hash(self, trigger):
        self.settings.disabled_assess_rules.discard("crypto-bad-mac")

        user_input = "def456"
        self.client.get("/vulnpy/hash/{}".format(trigger), {"user_input": user_input})

        validate_finding_in_context(
            self.request_context, "crypto-bad-mac", user_input, nondataflow=True
        )

    @pytest.mark.parametrize("trigger", ["exec", "eval", "compile"])
    def test_untrusted_code_exec(self, trigger):
        user_input = "1 + 2 + 3"
        self.client.get(
            "/vulnpy/unsafe_code_exec/{}".format(trigger), {"user_input": user_input}
        )

        validate_finding_in_context(
            self.request_context, "unsafe-code-execution", user_input
        )

    def test_raw_xss(self):
        user_input = "im an attack!"
        self.client.get("/vulnpy/xss/raw", {"user_input": user_input})

        validate_finding_in_context(self.request_context, "reflected-xss", user_input)

    @pytest.mark.parametrize(
        "trigger",
        ["xml-dom-pulldom-parsestring", "lxml-etree-fromstring", "xml-sax-parsestring"],
    )
    def test_xxe(self, trigger):
        user_input = "some input"

        try:
            self.client.get(
                "/vulnpy/xxe/{}".format(trigger), {"user_input": user_input}
            )
        except Exception:
            # this view function doesn't handle its own errors, so we do it here
            pass

        validate_finding_in_context(self.request_context, "xxe", user_input)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
# in a Flask environment, we should always have Werkzeug
from werkzeug.exceptions import NotFound

from contrast.agent.assess.rules.config import (
    FlaskHttpOnlyRule,
    FlaskSecureFlagRule,
    FlaskSessionAgeRule,
)
from contrast.agent.middlewares.route_coverage.flask_routes import (
    create_flask_routes,
    build_flask_route,
)
from contrast.wsgi.middleware import WSGIMiddleware
from contrast.utils.decorators import fail_safely


class FlaskMiddleware(WSGIMiddleware):
    def __init__(self, app):
        self.app = app

        self.config_rules = [
            FlaskSessionAgeRule(),
            FlaskSecureFlagRule(),
            FlaskHttpOnlyRule(),
        ]

        super(FlaskMiddleware, self).__init__(app.wsgi_app, app_name=app.name)

    def get_routes_on_first_request(self):
        """
        If an app is None, we won't do coverage.

        Flask requires there to be an app in the request context. There is only a
        context on init, which we use, and during the `self.wsgi_app` call. That is why
        flask.current_app.app_context doesn't work. There is nothing on the LocalProxy
        request context thread yet.
        """
        if self.app is None:
            return
        super(FlaskMiddleware, self).get_routes_on_first_request()

    @fail_safely("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        return create_flask_routes(self.app)

    @fail_safely("Unable to get Flask view func")
    def get_view_func(self, request):
        adapter = self.app.url_map.bind("empty")

        if None in (request, adapter):
            return None

        try:
            match = adapter.match(request.path, method=request.method)
        except NotFound:
            match = None

        func = None
        if match is not None:
            func = self.app.view_functions[match[0]]

        return func

    @fail_safely("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_flask_route(view_func.__name__, view_func)

    @fail_safely("Failed to run config scanning rules")
    def scan_configs(self):
        for rule in self.config_rules:
            rule.apply(self.app)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import importlib
import sys
import subprocess

try:
    import httplib as http_module
except ImportError:
    import http.client as http_module

import yaml.loader

import mock
from webtest import TestApp

from contrast.agent.policy import patch_manager
from contrast.agent.assess.utils import clear_properties
from contrast.agent.request_context import RequestContext
from contrast.extern import six
from contrast.test.settings_builder import SettingsBuilder
from contrast.agent.settings_state import SettingsState

SETTINGS_STATE_PATCH_LOC = "contrast.agent.middlewares.base_middleware.SettingsState"


class BaseContrastTest(object):
    @property
    def application_module_name(self):
        """
        The WSGI application module name, as a string.

        Example:
        ```
        # file: app/wsgi.py
        application = create_wsgi_application()
        ```

        Here, the `application` WSGI callable is defined in the app.wsgi module
        at app/wsgi.py. This means application_module_name should be "app.wsgi".
        """
        raise NotImplementedError

    @property
    def application_attribute_name(self):
        """
        The WSGI application attribute name, as a string.

        In the example above, application_attrubute_name should be "application".
        """
        raise NotImplementedError

    @property
    def middleware_request_context(self):
        """
        The name of the middleware containing the RequestContext instance that is
        used for requests to the application being tested. One day this will be
        unnecessary, as it will only be WSGIMiddleware.
        """
        raise NotImplementedError

    def setup_method(self):
        self.create_mocks()
        self.build_app()

    def teardown_method(self):
        self.settings_patch.stop()
        self.request_context_patch.stop()
        clear_properties()

        # this forces each new test to use a newly initialized application (& middleware)
        del sys.modules[self.application_module_name]

        # Deleting settings state instance caused stack overflow in pickle test
        SettingsState().heartbeat = None

    def create_mocks(self):
        self.settings = self.build_settings()
        mock_settings = mock.MagicMock(return_value=self.settings)
        self.settings_patch = mock.patch(SETTINGS_STATE_PATCH_LOC, mock_settings)

        self.request_context_patch = mock.patch(
            self.middleware_request_context, self.patch_request_context()
        )

        self.settings_patch.start()
        self.request_context_patch.start()

    def build_settings(self):
        """Override in subclasses to use different settings."""
        return SettingsBuilder().build()

    def patch_request_context(self):
        class PatchedRequestContext(RequestContext):
            def __init__(*args, **kwargs):
                super(PatchedRequestContext, args[0]).__init__(*args[1:], **kwargs)
                # This "self" belongs to the test class, not the request context
                self.request_context = args[0]

        return PatchedRequestContext

    def build_app(self):
        """
        Override to configure the app for each middleware. The goal of this configuration
        is to rebuild the WSGI application (and reinitialize the Contrast middleware) for
        each test. This way we avoid state bleeding across tests.
        """
        app_module = importlib.import_module(self.application_module_name)
        self.app = getattr(app_module, self.application_attribute_name)
        self.client = TestApp(self.app, lint=False)

    @classmethod
    def teardown_class(cls):
        """
        Patch locations that are used by both assess and protect tests need to be
        reversed here.
        """
        reverse_patch_module_names = [
            "os",
            "subprocess",
            "pickle",
            "io",
            "lxml.etree",
            "xml.dom.pulldom",
            "xml.sax",
            "urllib",
            "urllib2",
            "urllib.request",
            "__builtin__",
            "builtins",
        ]

        for module_name in reverse_patch_module_names:
            module = sys.modules.get(module_name)
            if module:
                patch_manager.reverse_patches_by_owner(module)

        reverse_patch_classes = [
            subprocess.Popen,
            http_module.HTTPConnection,
            yaml.loader.BaseLoader,
            yaml.loader.Loader,
            yaml.loader.UnsafeLoader,
            yaml.loader.FullLoader,
        ]

        try:
            import pymongo.collection

            reverse_patch_classes.append(pymongo.collection.Collection)
        except ImportError:
            pass

        for class_ in reverse_patch_classes:
            patch_manager.reverse_patches_by_owner(class_)

        # TODO: PYT-1188 everything below here is a workaround for the fact that reverse
        # patching also requires repatching

        from vulnpy.trigger import ssrf

        if six.PY3:
            import builtins
            import io
            from urllib.request import urlopen

            # needed because reverse_patches_by_owner doesn't run in PY3 for io because
            # io.open == builtins.open in PY3
            io.open = builtins.open

            ssrf.urlopen = urlopen
            ssrf.legacy_urlopen = urlopen
        else:
            from urllib2 import urlopen
            from urllib import urlopen as legacy_urlopen

            ssrf.urlopen = urlopen
            ssrf.legacy_urlopen = legacy_urlopen



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import threading
from contextlib import contextmanager
from collections import defaultdict


class ContextTracker(object):
    CURRENT_CONTEXT = "CURRENT_CONTEXT"
    PARENT_ATTR = "cs__parent_id"

    def __init__(self):
        self._tracker = defaultdict(dict)

    def get(self, key, default=None):
        try:
            return self._tracker[self.current_thread_id()][key]
        except KeyError:
            return default

    def clear(self):
        self._tracker.clear()

    def get_explicit_context(self, thread_id, key, default=None):
        """
        Can be used to retrieve a context from any thread, used when the current thread doesn't have a context
        but its 'parent' thread does
        """
        try:
            return self._tracker[thread_id][key]
        except KeyError:
            return default

    def set(self, key, value):
        self._tracker[self.current_thread_id()][key] = value

    def delete(self, key):
        current_thread_id = self.current_thread_id()

        if current_thread_id not in self._tracker:
            return

        del self._tracker[current_thread_id][key]

        if len(self._tracker[current_thread_id]) == 0:
            del self._tracker[current_thread_id]

    def set_current(self, value):
        self.set(self.CURRENT_CONTEXT, value)

    def delete_current(self):
        self.delete(self.CURRENT_CONTEXT)

    @contextmanager
    def lifespan(self, context):
        self.set_current(context)

        yield context

        self.delete_current()

    def current(self):
        context = self.get(self.CURRENT_CONTEXT)

        current_thread = threading.currentThread()

        if (
            context is None
            and hasattr(current_thread, self.PARENT_ATTR)
            and getattr(current_thread, self.PARENT_ATTR) in self._tracker
        ):
            return self.get_explicit_context(
                getattr(current_thread, self.PARENT_ATTR), self.CURRENT_CONTEXT
            )

        return context

    def current_thread_id(self):
        return threading.currentThread().ident



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import ast
import os
import sys
import traceback

from contrast.api.dtm_pb2 import StackTraceElement, TraceStack
from contrast.utils.decorators import fail_quietly
from contrast.utils.library_reader.library_reader import (
    get_active_library_names_from_pkg,
)
from contrast.utils.module_parser import get_ast_parsed_code

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


APPLIES_MARKER = "cs__"
PATCH_MARKER = "__cs"
PY_FILE_EXTENSION = ".py"
UTILS_MODULES = "contrast/utils"
NONETYPE = "NoneType"

CONTRAST_EXTENSIONS = ("contrast/assess_extensions", "contrast/patches")

DJANGO_EXCEPTION_PATH = "core/handlers/exception.py"
DJANGO_DEPRECATION_PATH = "utils/deprecation.py"

VIRTUAL_ENV = "VIRTUAL_ENV"
SITE_PACKAGES = "site-packages"


class StackTraceUtils(object):
    """
    Utility class to grab the latest stack frames based on depth and skip

    Can skip certain frames like our own

    Returns a list of StackTraceElement elements for the DTM
    """

    SORTED_SYS_PATH = sorted(sys.path, key=len, reverse=True)

    @staticmethod
    def build(
        skip=0,
        depth=10,
        ignore=False,
        ignore_string="/contrast/",
        class_lookup=False,
        for_trace=False,
    ):
        frames = traceback.extract_stack()
        frames.reverse()  # moves most recent to the front

        if ignore:
            # ignore frames with contrast or __cs
            def stack_filter(frame):
                if isinstance(frame, tuple):
                    return (
                        ignore_string not in frame[0]
                        and UTILS_MODULES not in frame[0]
                        and not any(
                            extension in frame[0] for extension in CONTRAST_EXTENSIONS
                        )
                        and not frame[2].startswith(APPLIES_MARKER)
                        and not frame[2].startswith(PATCH_MARKER)
                        and not frame[0].endswith(DJANGO_EXCEPTION_PATH)
                        and not frame[0].endswith(DJANGO_DEPRECATION_PATH)
                    )

                return (
                    ignore_string not in frame.filename
                    and UTILS_MODULES not in frame.filename
                    and not any(
                        extension in frame.filename for extension in CONTRAST_EXTENSIONS
                    )
                    and not frame.name.startswith(APPLIES_MARKER)
                    and not frame.name.startswith(PATCH_MARKER)
                    and not frame.filename.endswith(DJANGO_EXCEPTION_PATH)
                    and not frame.filename.endswith(DJANGO_DEPRECATION_PATH)
                )

            frames = [frame for frame in frames if stack_filter(frame)]

        max_frames = skip + depth
        return StackTraceUtils.to_element_list(
            frames[skip:max_frames], class_lookup, for_trace
        )

    @staticmethod
    def to_element_list(frames, class_lookup, for_trace):
        cached_classes = dict()
        return [
            y
            for y in [
                StackTraceUtils.to_element(x, class_lookup, cached_classes, for_trace)
                for x in frames
            ]
            if y is not None
        ]

    @staticmethod
    def to_element(string, class_lookup, cached_classes, for_trace):
        try:
            return StackTraceUtils._to_element(
                string, class_lookup, cached_classes, for_trace
            )
        except RuntimeError:
            return None

    @staticmethod
    def _to_element(summary, class_lookup, cached_classes, for_trace):
        if not summary:
            return None

        name_by_path = dict()

        if isinstance(summary, tuple):
            # in python 2 traceback returns a tuple
            path = summary[0]
            method = summary[2]
            line_number = summary[1]
        else:
            # python 3 is a FrameSummary
            path = summary.filename
            method = summary.name
            line_number = summary.lineno

        element = TraceStack() if for_trace else StackTraceElement()
        element.line_number = line_number

        element.file_name = StackTraceUtils.filename_formatter(path)

        if class_lookup:
            try:
                found_class = StackTraceUtils.find_class(
                    name_by_path, path, element.file_name, cached_classes, line_number
                )
            except Exception:
                found_class = None

            if found_class is None:
                # If the method for the view is not in a class, set the declaring class to NoneType
                element.declaring_class = NONETYPE
            elif not isinstance(found_class, str):
                element.declaring_class = found_class.__name__
            else:
                element.declaring_class = found_class
        else:
            element.declaring_class = path

        element.method_name = method

        return element

    @staticmethod
    def find_class(name_by_path, path, file_name, cached_classes, line_number):
        if not path:
            return None

        if path in name_by_path:
            return name_by_path[path]

        if not os.path.exists(path) or not os.path.isfile(path):
            return None

        parsed = get_ast_parsed_code(path)

        classes = [node for node in ast.walk(parsed) if isinstance(node, ast.ClassDef)]
        class_names = [node.name for node in classes]

        if file_name in cached_classes:
            cached_classes[file_name] = set(class_names).union(
                cached_classes[file_name]
            )
        else:
            cached_classes[file_name] = set(class_names)

        best_class = None

        if len(classes) > 0:
            for cls in classes:
                if line_number > cls.lineno:
                    for item in parsed.body:
                        if isinstance(item, ast.ClassDef) and item.lineno > cls.lineno:
                            best_class = item.name
                            break

        name_by_path[path] = best_class

        return best_class

    @staticmethod
    @fail_quietly("Unable to create file_name")
    def filename_formatter(file_name):
        if file_name.startswith("<frozen"):
            return file_name

        formatted = None

        if VIRTUAL_ENV in os.environ and SITE_PACKAGES in file_name:
            # agent is running in a python virtualenv
            python_lib = os.path.join(
                "lib", "python{}.{}".format(*sys.version_info[:2])
            )
            virtual_site = os.path.join(
                os.environ.get(VIRTUAL_ENV), python_lib, SITE_PACKAGES
            )
            formatted = file_name.replace(virtual_site, "")

        elif file_name.startswith(os.getcwd()):
            formatted = file_name.replace(os.getcwd(), "")

        else:
            for sys_path in StackTraceUtils.SORTED_SYS_PATH:
                if file_name.startswith(sys_path):
                    formatted = file_name.replace(sys_path, "")
                    break

        formatted = formatted or file_name

        return formatted.replace("/", ".").lstrip(".")

    @staticmethod
    def in_custom_code():
        current_stack_trace = StackTraceUtils.build(depth=4, ignore=True)
        for trace in current_stack_trace:
            if trace and isinstance(trace, StackTraceElement):
                file_name = trace.file_name
                if file_name.endswith(".py") or file_name.endswith(".pyc"):
                    file_name = file_name[:-3]
                if StackTraceUtils.is_custom_module(file_name):
                    return True
        return False

    @staticmethod
    def is_custom_module(file_name):
        """

        Given a file name, determines if the file is a custom module.

        If a library name is in the file_name, we can say that the file is part of that library
        Obviously this can go wrong if users are naming their custom modules library names but hopefully import errors
        deter people away from this.

        ex:
        file_name = 'flask.router.Router'
        active_library_names = ['flask', 'wsgi', 'sqlalchemy', 'git']
        returns False

        file_name = 'core.utils.docker_constants'
        active_library_names = ['flask', 'wsgi', 'sqlalchemy', 'git']
        return True

        :param file_name: string of file_name without file type (contrast.utils.stack_trace_utils)
        :return: True if no active libraries are in the file_name else False
        """
        active_libraries_names = [
            # exclude vulnpy - a purposely-vulnerable library
            lib
            for lib in get_active_library_names_from_pkg()
            if lib != "vulnpy"
        ]
        for library_name in active_libraries_names:
            if file_name.startswith(library_name):
                return False
        return True



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
# pylint: disable=too-many-lines
import copy
import os
import six  # pylint: disable=did-not-import-extern
import sys

import pytest
from app import settings
from app.urls import urlpatterns

import django
from django.utils.html import escape
from contrast.agent.protect.rule.deserialization_rule import Deserialization
from contrast.api.dtm_pb2 import AttackResult, ObservedRoute
from contrast.api.settings_pb2 import ProtectionRule
from contrast.test import mocks

from contrast.test.library_analysis import (
    assert_files_loaded,
    import_sample_package_onefile,
    TEST_MODULES_RELATIVE_IMPORT_DIST,
    TEST_MODULES_NAMESPACE_DIST,
    TEST_MODULE_ONEFILE,
    TEST_MODULES_MULTIPLE_TLDS_DIST,
)

from contrast.test.contract.findings import (
    assert_event_sources,
    param_is_tracked,
    remove_redos_findings,
    validate_finding_in_context,
    validate_source_finding_in_context,
    validate_header_sources,
    validate_cookies_sources,
)
from contrast.test.framework_test_utils import (
    assert_route_appended,
    assert_routes_count,
    current_routes_by_url,
    validate_observed_route,
)

from contrast.test.helper import (
    mock_build_update_messages,
    mock_send_messages,
)
from contrast.test.mongo import skip_no_mongo_db


from contrast.test.helper import python3_only

URLPATTERNS_COPY = copy.copy(urlpatterns)
DATA_DIR = os.path.join(settings.BASE_DIR, os.pardir, "data")

django_ver = django.__version__
django_2_0_1 = django_ver.startswith("2.0") or django_ver.startswith("2.1")

SOURCE_OPTIONS = (
    "parameter",
    "host",
    "port",
    "raw_uri",
    "scheme",
    "referer_header",
)

# full_path_info is only in newer versions of django.
# We do not test full_path* for POST requests since in this case the path does not
# contain a query, and so the entire path is actually sanitized.
GET_SOURCES = (
    SOURCE_OPTIONS
    + ("full_path",)
    + (("full_path_info",) if django.VERSION >= (2, 1) else ())
)
POST_SOURCES = ("files", "body",) + SOURCE_OPTIONS

re_methods = ["match", "search", "finditer", "findall", "sub", "subn", "split"]
re_pattern_methods = [method + "-compiled" for method in re_methods]

REDOS_TRIGGERS = re_methods + re_pattern_methods

if six.PY3:
    REDOS_TRIGGERS.extend(["fullmatch", "fullmatch-compiled"])


class DjangoTestLibraryAnalysisBuiltinImport(object):
    def test_relative_import_hook(self):
        response = self.client.get(
            "/import_package_with_relative_imports/", {"rm_sys_mod_entries": 1}
        )

        assert response.status_code == 200

        assert_files_loaded(
            self.request_context.activity, TEST_MODULES_RELATIVE_IMPORT_DIST
        )

    def test_namespace_sample_package_import_hook(self):
        response = self.client.get(
            "/import_namespace_package/", {"rm_sys_mod_entries": 1}
        )

        assert response.status_code == 200

        assert_files_loaded(self.request_context.activity, TEST_MODULES_NAMESPACE_DIST)

    def test_module_import_onefile(self):
        response = self.client.get(
            "/import_package_onefile/", {"rm_sys_mod_entries": 1}
        )

        assert response.status_code == 200

        assert_files_loaded(self.request_context.activity, TEST_MODULE_ONEFILE)

    def test_sample_module_multiple_tlds(self):
        response = self.client.get(
            "/import_sample_dist_multiple_tlds/", {"rm_sys_mod_entries": 1}
        )

        assert response.status_code == 200

        assert_files_loaded(
            self.request_context.activity, TEST_MODULES_MULTIPLE_TLDS_DIST
        )

    def test_module_already_imported(self):
        import_sample_package_onefile()

        response = self.client.get("/import_package_onefile/")

        assert response.status_code == 200

        assert len(self.request_context.activity.library_usages) == 0


class DjangoProtectNoSqliTestMixin(object):
    """
    Screener does not currently test nosqli protect so we test it more thoroughly.
    """

    def make_nosqli_request(self, method, param_val, with_kwarg):
        self.client.post(
            "/nosqli/",
            {
                "user_input": param_val,
                "method_to_test": method,
                "with_kwarg": with_kwarg,
            },
            status=403,
        )

    @skip_no_mongo_db
    @mocks.nosqli_input_analysis_mock
    @pytest.mark.parametrize("with_kwarg", [False, True])
    def test_nosqli_find_blocked(self, mocked_input_analysis, with_kwarg):
        param_val = "Record One"

        self.make_nosqli_request("find", param_val, with_kwarg)

        assert len(self.request_context.activity.results) == 1
        assert self.request_context.activity.results[0].response == AttackResult.BLOCKED
        assert self.request_context.activity.results[0].rule_id == "nosql-injection"
        assert self.request_context.activity.query_count == 1

    @skip_no_mongo_db
    @mocks.nosqli_input_analysis_mock
    @pytest.mark.parametrize("with_kwarg", [False, True])
    def test_nosqli_insert_one_blocked(self, mocked_input_analysis, with_kwarg):
        param_val = '{"title": "Record One"}'

        self.make_nosqli_request("insert_one", param_val, with_kwarg)

        assert len(self.request_context.activity.results) == 1
        assert self.request_context.activity.results[0].response == AttackResult.BLOCKED
        assert self.request_context.activity.results[0].rule_id == "nosql-injection"
        assert self.request_context.activity.query_count == 1

    @skip_no_mongo_db
    @mocks.nosqli_input_analysis_mock
    @pytest.mark.parametrize("with_kwarg", [False, True])
    def test_nosqli_insert_many_blocked(self, mocked_input_analysis, with_kwarg):
        param_val = "Record One"

        self.make_nosqli_request("insert_many", param_val, with_kwarg)

        assert len(self.request_context.activity.results) == 1
        assert self.request_context.activity.results[0].response == AttackResult.BLOCKED
        assert self.request_context.activity.results[0].rule_id == "nosql-injection"
        assert self.request_context.activity.query_count == 1

    @skip_no_mongo_db
    @mocks.nosqli_input_analysis_mock
    @pytest.mark.parametrize("with_kwarg", [False, True])
    def test_nosqli_update_blocked(self, mocked_input_analysis, with_kwarg):
        param_val = "Record One"

        self.make_nosqli_request("update", param_val, with_kwarg)

        assert len(self.request_context.activity.results) == 1
        assert self.request_context.activity.results[0].response == AttackResult.BLOCKED
        assert self.request_context.activity.results[0].rule_id == "nosql-injection"
        assert self.request_context.activity.query_count == 2

    @skip_no_mongo_db
    @mocks.nosqli_input_analysis_mock
    @pytest.mark.parametrize("with_kwarg", [False, True])
    def test_nosqli_delete_blocked(self, mocked_input_analysis, with_kwarg):
        param_val = "Record One"

        self.make_nosqli_request("delete", param_val, with_kwarg)

        assert len(self.request_context.activity.results) == 1
        assert self.request_context.activity.results[0].response == AttackResult.BLOCKED
        assert self.request_context.activity.results[0].rule_id == "nosql-injection"
        assert self.request_context.activity.query_count == 1


class DjangoProtectTestMixinBlock(DjangoProtectNoSqliTestMixin):
    """
    Provides test cases to be used across all django protect framework tests
    """

    PROTECT_RULES_MODE_BLOCK = [
        {
            "id": "cmd-injection",
            "name": "Command Injection",
            "mode": ProtectionRule.BLOCK,
        },
        {
            "id": "path-traversal",
            "name": "Path Traversal",
            "mode": ProtectionRule.BLOCK,
        },
        {
            "mode": ProtectionRule.BLOCK,
            "id": Deserialization.NAME,
            "name": "Untrusted Deserialization",
        },
        {"id": "nosql-injection", "name": "NosQLI", "mode": ProtectionRule.BLOCK},
    ]

    @pytest.mark.parametrize("path", ["os-system", "subprocess-popen"])
    def test_cmdi(self, path):
        self.client.get(
            "/vulnpy/cmdi/{}".format(path),
            {"user_input": "; echo /etc/passwd | nc"},
            status=403,
        )
        assert len(self.request_context.activity.results) == 1
        assert self.request_context.activity.results[0].response == AttackResult.BLOCKED
        assert self.request_context.activity.results[0].rule_id == "cmd-injection"

    @pytest.mark.parametrize("method_name", ["get", "post"])
    @pytest.mark.parametrize("trigger", ["pickle-load", "pickle-loads"])
    def test_pickle(self, trigger, method_name):
        param_val = "csubprocess\ncheck_output\n(S'ls'\ntR."

        params = dict(user_input=param_val)
        method = getattr(self.client, method_name)

        method("/vulnpy/deserialization/{}".format(trigger), params, status=403)

        assert len(self.request_context.activity.results) == 1
        assert self.request_context.activity.results[0].response == AttackResult.BLOCKED
        assert (
            self.request_context.activity.results[0].rule_id
            == "untrusted-deserialization"
        )


class DjangoAssessTestMixin(object):
    """
    Provides test cases to be used across all django assess framework tests
    """

    def test_track_django_sources(self):
        param_val = "; echo /etc/passwd | nc"
        response = self.client.get("/vulnpy/cmdi/os-system", {"user_input": param_val})

        assert response is not None
        assert param_is_tracked(param_val)

    @python3_only
    def test_non_utf8_decoding_cmdi(self):
        """
        param_val a non utf8 encoded string. This is the result of shift_jis encoding of
        '手袋'. This encoding is automatically done by chrome, not in our test client. So
        we manually provide the encoded bytes here.
        """
        trigger = "os-system"
        rule_id = "cmd-injection"
        # Couldn't rely on passing 手袋 in the request because the test client doesn't
        # use shift_jis to encode.
        param_val = b"\x8e\xe8\x91\xdc"
        # Since we fail to decode param_value using utf-8 we rely on replace error
        # handling to get this new string
        expected_value = "���"
        path = "/vulnpy/cmdi/{}".format(trigger)

        self.client.get(path, {"user_input": param_val})

        validate_finding_in_context(self.request_context, rule_id, expected_value)

    def test_xss_django_autoescape(self):
        """Autoescaping is enabled by default and should not lead to an xss finding"""
        param_val = "im an attack!"
        response = self.client.get("/xss-autoescape/", {"user_input": param_val})
        assert response.status_code == 200

        findings = remove_redos_findings(self.request_context.activity.findings, "xss")

        assert len(findings) == 0

    def test_xss_django_autoescape_safe(self):
        """
        Autoescaping is enabled but the value is marked as safe in the template
        which prevents it from being escaped. This should result in a finding.
        """
        param_val = "im an attack!"
        response = self.client.get("/xss-autoescape-safe/", {"user_input": param_val})
        assert response.status_code == 200

        validate_finding_in_context(self.request_context, "reflected-xss", param_val)

    def test_xss_django_no_autoescape(self):
        """Autoescaping is disabled so we should get a finding"""
        param_val = "im an attack!"
        response = self.client.get("/xss-no-autoescape/", {"user_input": param_val})
        assert response.status_code == 200

        validate_finding_in_context(self.request_context, "reflected-xss", param_val)

    def test_byte_file_contents_are_tracked(self):
        filename = "logo.png"
        file_path = os.path.join(DATA_DIR, filename)

        self.client.post("/file-contents/", upload_files=[("upload", file_path)])

        assert param_is_tracked(filename)

        contents = open(file_path, "rb").read()
        file_start = contents[:6]
        file_end = contents[-18:]

        # we cannot simply assert param_is_tracked(contents) because the contents
        # are split up into chunks in the string tracker and tracked in these chunks
        assert param_is_tracked(file_start)
        assert param_is_tracked(file_end)

    def test_uce_base64(self):
        # this string is "__import__('os').system('echo hacked')" b64 encoded
        encoded_input = "X19pbXBvcnRfXygnb3MnKS5zeXN0ZW0oJ2VjaG8gaGFja2VkJyk="
        response = self.client.get("/uce-base64/", {"user_input": encoded_input})

        assert response.status_code == 200
        findings = remove_redos_findings(
            self.request_context.activity.findings, "unsafe-code-execution"
        )

        assert len(findings) == 1

        assert findings[0].rule_id == u"unsafe-code-execution"
        assert len(findings[0].events[-1].args) == 1
        # Coincidentally, we base-64 encode our values to send to TS, so we can make
        # this assertion
        assert findings[0].events[-1].args[0].value == encoded_input

    @skip_no_mongo_db
    @pytest.mark.parametrize("with_kwarg", [False, True])
    def test_nosqli_rule_post_find(self, with_kwarg):
        param_val = "book title"
        self.client.post(
            "/nosqli/",
            {
                "user_input": param_val,
                "method_to_test": "find",
                "with_kwarg": with_kwarg,
            },
        )

        validate_finding_in_context(self.request_context, "nosql-injection", param_val)

    @skip_no_mongo_db
    @pytest.mark.parametrize("with_kwarg", [False, True])
    def test_nosqli_rule_post_insert_one(self, with_kwarg):
        values = ("Record One", "PyMongo is fun!", "Dani")
        param_val = (
            '{"title": "'
            + values[0]
            + '", "content": "'
            + values[1]
            + '", "author": "'
            + values[2]
            + '"}'
        )

        self.client.post(
            "/nosqli/",
            {
                "user_input": param_val,
                "method_to_test": "insert_one",
                "with_kwarg": with_kwarg,
            },
        )

        # pymongo rewrites the query quite a bit before trigger time, so we can't assert
        # that the entire query we pass in makes it to the trigger. It's still a pretty
        # good check that one of our query values makes it to the trigger instead. Also,
        # older versions of python have different behavior, so we skip the check when
        # appropriate by setting user_input to an empty string.
        validate_finding_in_context(
            self.request_context,
            "nosql-injection",
            "" if sys.version_info < (3, 6) else values[0],
        )

    @skip_no_mongo_db
    @pytest.mark.parametrize("with_kwarg", [False, True])
    def test_nosqli_rule_post_insert_many(self, with_kwarg):
        example_title = "Record One"
        param_val = (
            '{{"title": "'
            + example_title
            + '", "content": "PyMongo is fun!", "author":'
            ' "Dani"},{"title": "Record Two", "content": "PyMongo is alright",'
            ' "author": "Dani"}}'
        )

        self.client.post(
            "/nosqli/",
            {
                "user_input": param_val,
                "method_to_test": "insert_many",
                "with_kwarg": with_kwarg,
            },
        )

        validate_finding_in_context(
            self.request_context,
            "nosql-injection",
            "" if sys.version_info < (3, 6) else example_title,
        )

    @skip_no_mongo_db
    def test_nosqli_rule_post_update(self):
        param_val = "New Title"

        self.client.post(
            "/nosqli/", {"user_input": param_val, "method_to_test": "update"},
        )

        validate_finding_in_context(
            self.request_context, "nosql-injection", param_val, trigger_arg=1
        )

    @skip_no_mongo_db
    def test_nosqli_rule_post_delete(self):
        param_val = (
            '{"title": "Record One", "content": "PyMongo is fun!", "author": "Dani"}'
        )

        self.client.post(
            "/nosqli/", {"user_input": param_val, "method_to_test": "delete"},
        )

        validate_finding_in_context(self.request_context, "nosql-injection", param_val)

    def test_xss_escaped_no_vuln(self):
        param_val = "<script>alert(1)</script>"
        response = self.client.get(
            "/xss/", {"user_input": param_val, "django_escape": "True"},
        )

        assert response is not None
        assert param_is_tracked(param_val)
        assert param_is_tracked(escape(param_val))

        findings = remove_redos_findings(self.request_context.activity.findings, "xss")

        assert len(findings) == 0

    def test_xss_jinja(self):
        param_val = "im an attack!"
        self.client.get("/xss-jinja/", {"user_input": param_val})

        validate_finding_in_context(self.request_context, "reflected-xss", param_val)

    def test_xss_mako(self):
        if django.VERSION[:1] == (3,):
            pytest.xfail("djangomako is broken in Django 3")

        param_val = "im an attack!"
        self.client.get("/xss-mako/", {"user_input": param_val})

        validate_finding_in_context(self.request_context, "reflected-xss", param_val)

    @pytest.mark.parametrize("route", ["/vulnpy/xss/raw", "/raw-xss-streaming/"])
    def test_xss_raw(self, route):
        param_val = "im an attack!"

        response = self.client.get(route, {"user_input": param_val})
        assert response.status_code == 200

        if "streaming" in route:
            # Make sure that our middleware doesn't consume streaming content
            content = response.body
            assert content == b"<p>Looks like xss: im an attack!</p>"
        else:
            assert "<p>XSS: {}</p>".format(param_val) in str(response.body)

        validate_finding_in_context(self.request_context, "reflected-xss", param_val)

    def test_xss_csv_does_not_send_finding(self):
        param_val = "im an attack!"
        response = self.client.get("/xss-csv/", {"user_input": param_val})

        assert response.status_code == 200

        findings = remove_redos_findings(self.request_context.activity.findings, "xss")

        assert len(findings) == 0

    def test_cmdi_from_filename_and_contents(self):
        filename = "testfile.txt"
        file_path = os.path.join(DATA_DIR, filename)

        self.client.post("/cmdi-file/", upload_files=[("upload", file_path)])

        validate_finding_in_context(self.request_context, "cmd-injection", filename)

        file_contents = (
            b"This file contains extremely dangerous data. Use at your own risk!\n"
        )
        param_is_tracked(file_contents)

    def test_stream_source(self):
        response = self.client.get("/stream-source/")
        assert response.status_code == 200

        # This endpoint is pretty contrived and marks a stream as a source directly.
        # As such, we don't have any user input for validation
        validate_finding_in_context(
            self.request_context, "untrusted-deserialization", "see comment above"
        )

    @pytest.mark.parametrize(
        "trigger_method", REDOS_TRIGGERS,
    )
    def test_redos(self, trigger_method):
        user_input = "anything"

        response = self.client.get(
            "/vulnpy/redos/re-{}".format(trigger_method), {"user_input": user_input},
        )

        assert response.status_code == 200

        regex = "((a)+)+"
        if (
            os.environ.get("USE_WSGI")
            and len(self.request_context.activity.findings) == 2
        ):
            # we see multiple redos findings with django WSGI
            # We only care about the second one
            self.request_context.activity.findings.pop(0)
        validate_finding_in_context(self.request_context, "redos", regex)

    @pytest.mark.parametrize("permanent_redirect", [True, False])
    def test_unvalidated_redirect(self, permanent_redirect):
        """
        By passing a permanent redirect bool flag, we can test that both
        HttpResponsePermanentRedirect and HttpResponseRedirect are patched
        because they inherit from HttpResponseRedirectBase
        """
        redirect_route = "/vulnpy"
        self.client.get(
            "/unvalidated-redirect/",
            {"user_input": redirect_route, "permanent": permanent_redirect},
        )

        validate_finding_in_context(
            self.request_context, "unvalidated-redirect", redirect_route
        )

    def test_preflight_hash_same_request(self):
        """
        Two requests that look exactly the same should generate the same hash
        """
        # This test will pass in isolation on django-1.10.8, but when the full test
        # suite runs, it will not. The findings are both present, but the first one
        # has way more events than the second one for some reason. It seems like
        # something in the test environment is polluting this test state somehow.
        if django.VERSION[:2] == (1, 10):
            pytest.xfail("Known weirdness with events on django-1.10.8 framework test")

        self.client.get("/vulnpy/cmdi/os-system", {"user_input": "first attack"})

        findings = remove_redos_findings(self.request_context.activity.findings, "cmdi")
        assert len(findings) == 1
        first_finding = findings[0]

        self.client.get("/vulnpy/cmdi/os-system", {"user_input": "second attack"})
        findings = remove_redos_findings(self.request_context.activity.findings, "cmdi")
        assert len(findings) == 1
        second_finding = findings[0]

        assert second_finding is not first_finding
        assert second_finding.hash_code == first_finding.hash_code
        assert second_finding.preflight == first_finding.preflight

    def test_preflight_hash_different_requests(self):
        """
        Two different requests that find the same vuln should have different hashes

        In this case, the two requests are the same except one uses GET and the
        other uses POST. Since the HTTP method contributes to the hash, we
        should get different preflight values.
        """
        param_val = "im an attack!"
        self.client.get("/vulnpy/cmdi/subprocess-popen", {"user_input": param_val})

        findings = remove_redos_findings(self.request_context.activity.findings, "cmdi")
        assert len(findings) == 1

        first_finding = findings[0]

        self.client.post("/vulnpy/cmdi/subprocess-popen", {"user_input": param_val})
        findings = remove_redos_findings(self.request_context.activity.findings, "cmdi")
        assert len(findings) == 1
        second_finding = findings[0]

        assert second_finding is not first_finding
        assert second_finding.hash_code != first_finding.hash_code
        assert second_finding.preflight != first_finding.preflight

    def test_preflight_hash_same_request_different_vulns(self):
        """
        The same request that generates two different vulns should have different hashes
        """
        param_val = "im an attack!"
        self.client.get("/two-vulns/", {"user_input": param_val})

        findings = remove_redos_findings(self.request_context.activity.findings, "xss")
        assert len(findings) == 2

        first_finding = self.request_context.activity.findings[0]
        second_finding = self.request_context.activity.findings[1]

        assert second_finding.hash_code != first_finding.hash_code
        assert second_finding.preflight != first_finding.hash_code

    @pytest.mark.xfail(django.VERSION[:2] == (1, 9), reason="TODO: PYT-1050")
    @pytest.mark.parametrize(
        "test_type",
        [
            "create_method",
            "init_direct",
            "filter",
            # TODO: PYT-583 still need to test multi-trigger / multi findings
            "multi-column",
        ],
    )
    @pytest.mark.django_db(transaction=False)
    def test_stored_xss(self, test_type):
        param_val = "hello"
        self.client.get(
            "/stored-xss/", {"user_input": param_val, "test_type": test_type}
        )

        findings = self.request_context.activity.findings

        validate_finding_in_context(self.request_context, "reflected-xss", param_val)

        assert_event_sources(findings, "database", {"database": "TAINTED_DATABASE"})

    @pytest.mark.parametrize("module_name", ["xml", "lxml"])
    def test_xpath_injection(self, module_name):
        query = ".//*[@name='whatever']"
        self.client.get("/xpath-injection/", dict(query=query, module=module_name))

        validate_finding_in_context(self.request_context, "xpath-injection", query)

    @pytest.mark.django_db(transaction=False)
    @pytest.mark.parametrize("method", ["get", "post"])
    def test_sqli_via_django_orm(self, method):
        param_val = "doesnt matter"
        getattr(self.client, method)("/sqli/", {"user_input": param_val})

        validate_finding_in_context(self.request_context, "sql-injection", param_val)


class DjangoAssessTestNondataflowMixin(object):
    def test_crypto_bad_ciphers(self):
        self.client.get("/crypto-bad-ciphers/")

        validate_finding_in_context(
            self.request_context, "crypto-bad-ciphers", None, nondataflow=True
        )


class DjangoRouteCoverageTestMixin(object):
    @mock_build_update_messages
    def test_finds_app_routes(self, build_update_message):
        self.client.get("/vulnpy")

        routes = build_update_message.call_args[0][0]
        cmdi_routes = current_routes_by_url(routes, "vulnpy/cmdi/os-system")
        assert_routes_count(cmdi_routes, 2)

    @mock_build_update_messages
    def test_reports_dynamic_route_in_discovery(self, build_update_message):
        """
        Tests that a dynamically-created route is found because the first request
        made is to the endpoint that creates this route. The dynamically-created
        route is not visited.
        """
        new_view = "new_dynamic"
        original_url_names = [x.name for x in URLPATTERNS_COPY[1:]]
        assert new_view not in original_url_names

        self.client.get("/dynamic_url_view/", {"user_input": new_view})

        routes = build_update_message.call_args[0][0]
        dynamic_routes = current_routes_by_url(routes, new_view)
        assert_routes_count(dynamic_routes, 2)

    @mock_send_messages
    @mock_build_update_messages
    def test_reports_dynamic_route_by_observed_route(
        self, build_update_message, send_messages
    ):
        """
        Tests that a dynamically-created route is found because it is visited. The
        endpoint that created this route was not the first request to the app.
        """
        new_view = "other_dynamic"
        original_url_names = [x.name for x in URLPATTERNS_COPY[1:]]
        assert new_view not in original_url_names
        self.client.get("/")

        routes = build_update_message.call_args[0][0]
        current_routes = current_routes_by_url(routes, new_view)
        # new_dynamic has not been created so it is not found
        assert_routes_count(current_routes, 0)

        assert len(send_messages.call_args[0][0]) == 4

        self.client.get("/dynamic_url_view/?user_input=" + new_view)

        assert len(send_messages.call_args[0][0]) == 3

        new_path = "/" + new_view + "/"
        self.client.get(new_path)
        call_args = send_messages.call_args[0][0]
        assert len(call_args) == 3

        observed_route = call_args[2]
        assert isinstance(observed_route, ObservedRoute)
        assert observed_route.url == new_path
        assert (
            observed_route.signature
            == "contrast.test.framework_test.django_views.other_dynamic(request,)"
        )

    def test_route_appended_to_xss_vuln(self):
        param_val = "im an attack!"
        response = self.client.get("/xss/", {"user_input": param_val})

        assert response
        assert len(self.request_context.activity.findings) > 0
        assert_route_appended(self.request_context.activity.findings, "xss", "GET")

    @pytest.mark.django_db(transaction=False)
    def test_observed_route_appended(self):
        param_val = "im an attack!"
        param_name = "user_input"
        url = "/sqli"
        response = self.client.get(url, {param_name: param_val})

        assert response
        findings = remove_redos_findings(
            self.request_context.activity.findings, "sql-injection"
        )

        assert len(findings) == 1
        validate_observed_route(
            self.request_context, "GET", url[1:], param_name, "PARAMETER"
        )

    @pytest.mark.parametrize("setdefault", ["False", "True"])
    def test_trust_boundary_violation(self, setdefault):
        user_input = "does not matter"

        self.client.get(
            "/trust-boundary-violation/",
            {"user_input": user_input, "setdefault": setdefault},
        )
        findings = remove_redos_findings(
            self.request_context.activity.findings, "trust-boundary-violation"
        )

        # the wsgi middleware is early enough to see a crypto-bad-mac in an earlier middleware
        # the django middleware is last in the middleware stack, so it misses this vulnerability
        if os.environ.get("USE_WSGI"):  # TODO: PYT-1102
            assert len(findings) == 2
        else:
            validate_finding_in_context(
                self.request_context,
                "trust-boundary-violation",
                user_input,
                trigger_arg=1,
            )

    @pytest.mark.skipif(
        not (django.VERSION[:2] == (1, 9) and os.environ.get("USE_WSGI")),
        reason="see comment; only applies to django 1.9 + djangoWSGI",
    )
    def test_django19_source_tagging_196(self):
        """
        There's an unfortunate interaction in django 1.9 using the WSGI-based mdw:

        - in the django 1.9 implementation, only the url-encoded parts of the string are
          explicitly urldecoded. This is unfortunate for us because other parts of the
          string are still tagged as encoded (safe).
        - if the string doesn't have any characters that require url encoding / decoding
          then the entire string is still tagged with the encoded tag
        - this means that for vulnerabilities to trigger with query parameters in django
          1.9, they must have at least one character that is URL-decoded

        This test demonstrates the behavior so that we don't forget about it. It's not
        correct behavior, but fixing it would be quite difficult.
        """
        self.client.get(
            "/xss/", {"user_input": "noUrlDecodableCharactersInThisString"},
        )
        findings = remove_redos_findings(
            self.request_context.activity.findings, "trust-boundary-violation"
        )
        assert len(findings) == 0


class DjangoAssessTestDynamicSourcesMixin(object):

    ATTACK_VALUE = "im an attack!"
    DYNAMIC_SOURCE_RULE_ID = "cmd-injection"

    def set_dangerous_cookie(self):
        self.client.set_cookie("user_input", self.ATTACK_VALUE)

    def test_xss_from_cookie(self):
        self.set_dangerous_cookie()
        self.client.get("/xss-cookie/")

        findings = remove_redos_findings(self.request_context.activity.findings, "xss")
        assert len(findings) == 0

    @pytest.mark.parametrize("source", GET_SOURCES)
    def test_all_get_sources(self, source):

        self.set_dangerous_cookie()
        self.client.get(
            "/dynamic-sources/",
            {"user_input": self.ATTACK_VALUE, "source": source},
            headers={"Referer": "www.python.org"},
        )

        validate_source_finding_in_context(
            self.request_context, self.DYNAMIC_SOURCE_RULE_ID, source,
        )

    @pytest.mark.parametrize("source", POST_SOURCES)
    def test_all_post_sources(self, source):

        self.set_dangerous_cookie()
        if source == "files":
            file_path = os.path.join(DATA_DIR, "testfile.txt")
            self.client.post(
                "/dynamic-sources/",
                {"source": source},
                upload_files=[("user_input", file_path)],
                headers={"Referer": "www.python.org"},
            )

        else:
            self.client.post(
                "/dynamic-sources/",
                {"user_input": self.ATTACK_VALUE, "source": source},
                headers={"Referer": "www.python.org"},
                # **{source: self.ATTACK_VALUE}
            )

        validate_source_finding_in_context(
            self.request_context, self.DYNAMIC_SOURCE_RULE_ID, source,
        )

    def test_cookie_get(self):
        """Cookies shouldn't trigger xss, so we need a separate cookie source test"""
        self.set_dangerous_cookie()
        self.client.get("/cookie-source/")
        validate_cookies_sources(self.request_context.activity.findings)

    def test_cookie_post(self):
        """Cookies shouldn't trigger xss, so we need a separate cookie source test"""
        self.set_dangerous_cookie()
        self.client.post("/cookie-source/")
        validate_cookies_sources(self.request_context.activity.findings)

    @pytest.mark.parametrize(
        "route,source_name",
        [("/header-source/", "header"), ("/header-key-source/", "header_key")],
    )
    @pytest.mark.parametrize("method_name", ["get", "post"])
    @pytest.mark.parametrize("use_environ", [False, True])
    def test_header_source(self, use_environ, method_name, route, source_name):
        method = getattr(self.client, method_name)
        method(route, headers={"Test-Header": "whatever"})

        validate_header_sources(self.request_context.activity.findings, source_name)

    @pytest.mark.parametrize("method_name", ["get", "post"])
    def test_http_method_not_source(self, method_name):
        method = getattr(self.client, method_name)
        method("/dynamic-sources/", dict(source="http_method"))

        findings = remove_redos_findings(self.request_context.activity.findings, "xss")
        assert len(findings) == 0



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import os
import sys

from contrast.agent import scope
from contrast.agent.assess.policy.source_node import SourceNode
from contrast.agent.assess.policy.source_policy import cs__apply_source
from contrast.agent.assess.rules.config import (
    DjangoHttpOnlyRule,
    DjangoSecureFlagRule,
    DjangoSessionAgeRule,
)

from contrast.agent.middlewares.base_middleware import BaseMiddleware
from contrast.agent.middlewares.environ_tracker import (
    track_cookie_sources,
    track_environ_sources,
)
from contrast.agent.middlewares.route_coverage.django_routes import (
    create_django_routes,
    build_django_route,
)
from contrast.utils.decorators import fail_safely
from contrast.utils.exceptions.wrong_django_middleware_exception import (
    WrongDjangoMiddlewareException,
)

FILE_NODE_DICT = {
    "module": "django.http.request",
    "class_name": "HttpRequest",
    "instance_method": False,
    "method_name": "FILES",
    "target": "RETURN",
    "type": "PARAMETER",
}

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


class DjangoBaseMiddleware(BaseMiddleware):
    """
    Base class for all django middlewares.
    """

    def __init__(self):
        self.app_name = self.get_app_name()

        self.config_rules = [
            DjangoSessionAgeRule(),
            DjangoSecureFlagRule(),
            DjangoHttpOnlyRule(),
        ]

        super(DjangoBaseMiddleware, self).__init__()

    def get_app_name(self):
        try:
            from django.conf import settings

            wsgi_application = settings.WSGI_APPLICATION

            return wsgi_application.split(".")[0]
        except:
            return "root"

    @fail_safely("Unable to get Django view func")
    def get_view_func(self, request):
        match = request.resolver_match
        if not match:
            return None

        return match.func

    @fail_safely("Unable to build route", return_value="")
    def build_route(self, view_func, url):
        return build_django_route(view_func)

    @fail_safely("Unable to get route coverage", return_value={})
    def get_route_coverage(self):
        """
        Route Coverage is the Assess feature that looks for routes generally defined
        in Django apps in a file like urls.py
        """
        return create_django_routes()

    def generate_security_exception_response(self):
        from django.http.response import HttpResponseForbidden

        return HttpResponseForbidden(content=self.OVERRIDE_MESSAGE)

    def call_with_agent(self, context, request):
        self.log_start_request_analysis(request.path)
        context.timer.set_start("total")

        # For now, django does not track the wsgi.input stream since it's not a true IO
        # stream object and we have other ways of getting the sources we need. No need
        # to track request.META since it is equivalent to request.environ.
        track_environ_sources("django", context, request.environ, skip_wsgi_input=True)
        track_cookie_sources("django", context, request.COOKIES)
        with scope.contrast_scope():
            self.track_files(context, request)

    def track_files(self, context, request):
        """
        Add uploaded files as sources
        """
        if not self.settings.is_assess_enabled():
            return

        node = SourceNode.from_dict("django", FILE_NODE_DICT)
        for name, fileobj_list in request.FILES.lists():
            cs__apply_source(context, node, name, request, name, (), {})
            for fileobj in fileobj_list:
                if hasattr(fileobj, "file") and hasattr(fileobj.file, "cs__source"):
                    fileobj.file.cs__source = True
                    fileobj.file.cs__source_type = "MULTIPART_CONTENT_DATA"
                    fileobj.file.cs__source_tags = ["UNTRUSTED", "CROSS_SITE"]

                cs__apply_source(
                    context, node, fileobj.name, request, fileobj.name, (), {}
                )

    def validate_middleware(self):
        """
        Verify that the middleware is being used properly (legacy -> MIDDLEWARE_CLASSES, standard -> MIDDLEWARE).
        If it is not, raise a RuntimeError with a helpful error message.
        """
        from django.conf import settings

        middleware_name = ".".join([self.__class__.__module__, self.__class__.__name__])
        middlewares = getattr(settings, self.setting_name, []) or []
        if middleware_name not in middlewares:
            # isinstance(self, LegacyDjangoMiddleware) would require a circular import
            raise WrongDjangoMiddlewareException("legacy" in self.__class__.__module__)

    @staticmethod
    @fail_safely(
        "Unable to access request body, likely due to an exhaustive read by an earlier"
        " middleware. We suggest moving the contrast middleware higher up in the"
        " middleware stack."
    )
    def get_body(request):
        return request.body

    @fail_safely("Failed to run config scanning rules")
    def scan_configs(self):
        """
        Run config scanning rules for assess

        Overridden from base class; gets called from base class
        """
        from django.conf import settings as app_settings

        app_config_module_name = os.environ.get("DJANGO_SETTINGS_MODULE")
        if not app_config_module_name:
            logger.warning("Unable to find Django settings for config scanning")
            return

        app_config_module = sys.modules.get(app_config_module_name)
        if not app_config_module:
            logger.warning("Django settings module not loaded; can't scan config")
            return

        for rule in self.config_rules:
            rule.apply(app_settings, app_config_module)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
from collections import OrderedDict
import inspect
import sys

from contrast.utils.decorators import fail_safely
from contrast.extern.six import iteritems, PY2

from contrast.agent.policy import patch_manager
from contrast.utils.ignored_modules import (
    ALL_MODULES_TO_IGNORE,
    MODULES_TO_IGNORE_PREFIXES,
)
from contrast.utils.object_share import ObjectShare

from contrast.extern import structlog as logging

logger = logging.getLogger("contrast")


def module_is_ignored(module_name):
    return module_name in ALL_MODULES_TO_IGNORE or module_name.startswith(
        MODULES_TO_IGNORE_PREFIXES
    )


def call_patch(
    new_function, original_function, patch_policy, *orig_func_args, **orig_func_kwargs
):
    return new_function(
        original_function, patch_policy, *orig_func_args, **orig_func_kwargs
    )


def curry(new_function, original_func, patch_policy):
    def _curried(*orig_func_args, **orig_func_kwargs):
        return call_patch(
            new_function,
            original_func,
            patch_policy,
            *orig_func_args,
            **orig_func_kwargs
        )

    return _curried


def patch_cls_or_instance(
    patch_target,
    method_name,
    new_function=None,
    patch_policy=None,
    static_method=False,
    class_method=False,
    assess_patch=False,
):
    """
    Apply a patch to given module, class, or object instance

    :param patch_target: The class, module, or object instance to be patched
    :param method_name: The name of the method (or function) that is being replaced
    :param patch_policy: patch policy location instance
    :param new_function: The function that will replace the original function
    :param static_method: Indicates whether method being replaced is a staticmethod
    :param class_method: Indicates whether method being replaced is a classmethod
    """
    cs_method_name = (
        ObjectShare.CONTRAST_ASSESS_METHOD_START + method_name
        if assess_patch
        else method_name
    )

    original_func = getattr(patch_target, cs_method_name)
    if patch_manager.is_patched(original_func):
        # TODO: PYT-692 investigate unexpected patching
        logger.debug(
            "WARNING! Method %s is not getting patched - already patched", method_name
        )
        return

    curried = curry(new_function, original_func, patch_policy)

    if static_method:
        curried_function = staticmethod(curried)
    elif class_method:
        curried_function = classmethod(curried)
    elif callable(new_function) or hasattr(new_function, "__call__"):
        if isinstance(original_func, property):
            curried_function = property(
                fget=curry(new_function, method_name, patch_policy),
                fset=original_func.fset,
                fdel=original_func.fdel,
            )
        else:
            curried_function = curried
    else:
        logger.debug(
            "WARNING! Method %s is not getting patched - unrecognized object type: %s",
            method_name,
            type(original_func),
        )
        return

    patch_manager.patch(patch_target, method_name, curried_function)


def patch_property(
    class_to_patch, property_name, new_property_method, patch_policy=None
):
    """
    Patch properties which require both fget and fset to be set.
    This currently supports the case for the DjangoORM. Column name attributes
    for the model that defines database operations operate as properties.
    While in patch_cls_or_instance we handle properties, here we handle this special
    case that fails if fset is not defined.

    Note the definition of fset is the most basic definition of a setter, which is
    sufficient to make the getter/setter implementation of the property patching work.
    """
    original_prop = getattr(class_to_patch, property_name)
    if patch_manager.is_patched(original_prop):
        return

    fget = curry(new_property_method, property_name, patch_policy)

    def fset(cls_instance, value):
        # while good instinct would lead us to use setattr here instead of __dict__,
        # doing so does not work because we are in fact within a setter!
        cls_instance.__dict__[property_name] = value

    new_property = property(fget=fget, fset=fset)
    patch_manager.patch(class_to_patch, property_name, new_property)


CONTRAST_MODULE = "contrast."
INTELLIJ_MODULE = "pydevd"


def get_loaded_modules(use_for_patching=False):
    """
    Retrieves and filters all loaded modules

    The parameter `use_for_patching` indicates that this function is being
    called to enable patching. In this case the modules are sorted (to
    provide deterministic behavior) and also the modules_to_ignore list is used.

    NOTE: This method gets called multiple times during the course of agent
    initialization. Ideally it would be called only once for PERF optimization,
    but because sys.modules is global to all threads, we can't guarantee its contents
    will be the same and that a race condition won't happen which would add modules
    across different threads.

    :return: dict of name and module as value
    """
    if not use_for_patching:
        # Have to make a copy of sys.modules in order to avoid RuntimeError: dictionary changed size during iteration
        return {k: v for k, v in iteritems(dict(sys.modules)) if inspect.ismodule(v)}

    filtered = OrderedDict()
    filtered.update(
        dict(
            (name, module)
            for name, module in sorted(iteritems(sys.modules))
            if inspect.ismodule(module)
            and not module_is_ignored(name)
            and not is_so_module(module)
        )
    )

    return filtered


def is_so_module(module):
    """
    Return True if module is an .so file, such as
    ".../readline.cpython-38-darwin.so"

    :param module: python module object
    :return: bool
    """
    if not hasattr(module, "__file__") or module.__file__ is None:
        return False

    return module.__file__.endswith(".so")


def is_patchable(obj):
    if inspect.ismodule(obj):
        return False
    if inspect.isclass(obj):
        return True

    # cython methods look like unpatchable builtins, but they can be patched normally
    # an example of this is lxml.etree.fromstring
    # for additional info, see https://groups.google.com/forum/#!topic/cython-users/v5dXFOu-DNc
    is_unpatchable_builtin_method = inspect.ismethoddescriptor(
        obj
    ) and not obj.__class__.__name__.startswith("cython")

    return inspect.isroutine(obj) and not is_unpatchable_builtin_method


@fail_safely("Unable to repatch single module", log_level="exception")
def repatch_module(module):
    """Repatch a single module. See docstring for repatch_imported_modules"""

    # dict.items() returns a copy in Python 2. Didn't want to make another copy so doing list(...)
    # in a separate PY3 block. In Python 3 dict.items() returns a dict_view not a copy.
    module_attrs = vars(module).items() if PY2 else list(vars(module).items())

    for attr_name, attr in module_attrs:
        try:
            if not is_patchable(attr):
                continue
        except Exception as e:
            logger.debug(
                "exception occurred while checking whether to patch %s in %s",
                attr_name,
                module.__name__,
                exc_info=e,
            )
            continue

        if not patch_manager.has_associated_patch(attr):
            continue

        logger.debug("applying repatch to %s in %s", attr_name, module.__name__)
        patch_manager.patch(module, attr_name)


@fail_safely("Unable to patch previously imported modules", log_level="exception")
def repatch_imported_modules():
    """
    Applies patches to modules that were already imported prior to agent startup

    Here's the problem: our patches don't get applied until after our
    middleware class is initialized. At this point it's likely that most (or
    all) application modules will have already been imported.

    If we patch the function `foo.bar.baz`, and an application module that was
    loaded prior to our patches imports it as `from foo.bar import baz`, then
    our patch will have no effect within that application module. This is
    because the application module has a reference to the *original* function,
    and that reference remains unchanged even after we apply a patch to the
    `foo.bar` module.

    On the other hand, if the application imports it as `from foo import bar`
    and uses it as `bar.baz()`, then our patches will work just fine. In this
    case, the application module has a reference to the *module itself*, which
    is where we apply our patch. This means that when the application calls
    `bar.baz()`, it will be calling the updated (patched) function.

    Incidentally, if the application imports as `from foo.bar import baz`, but
    this module is not loaded until *after* our patches have been applied, our
    patch will be effective. However, we have no control over the order of
    imports in an application.

    This function is designed to remedy the former case in order to make sure
    that our patches are effective regardless of how they are imported or the
    order in which they are imported by the application.

    Prior to calling this function, we make a record of every function that
    gets patched. After all patches are applied, this function iterates
    through all imported modules, which includes all modules that may have been
    imported before our patches were applied. We look for any instances of the
    original functions that need to be patched, and we replace them with the
    patches in those modules.
    """
    for _, module in get_loaded_modules(use_for_patching=True).items():
        repatch_module(module)



# -*- coding: utf-8 -*-
# Copyright © 2021 Contrast Security, Inc.
# See https://www.contrastsecurity.com/enduser-terms-0317a for more details.
import copy


def safe_copy(value):
    """
    Return a safe copy of a value

    :param value: to be copied
    :return: copied value if no exception
    """
    try:
        return copy.copy(value)
    except:
        return value
